"""
Tests for grammar parser.
"""
__copyright__ = "Copyright (C) 2015-2016  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
import unittest
import inspect
import textwrap
import re
from decimal import Decimal
from unittest import mock

from beancount.core.number import D
from beancount.core.number import ZERO
from beancount.core.number import MISSING
from beancount.core.amount import from_string as A
from beancount.core.amount import Amount
from beancount.core.position import CostSpec
from beancount.parser import grammar
from beancount.parser import parser
from beancount.parser import lexer
from beancount.core import data
from beancount.core import amount
from beancount.utils import test_utils
from beancount.parser import cmptest


def check_list(test, objlist, explist):
    """Assert the list of objects against the expected specification.

    Args:
      test: the instance of the test object, used for generating assertions.
      objlist: the list of objects returned.

      explist: the list of objects expected. 'explist' can be an integer, to
               check the length of the list; if it is a list of types, the types
               are checked against the types of the objects in the list. This is
               meant to be a convenient method.
    """
    if isinstance(explist, int):
        test.assertEqual(explist, len(objlist))
    elif isinstance(explist, (tuple, list)):
        test.assertEqual(len(explist), len(objlist))
        for obj, exp in zip(objlist, explist):
            test.assertIsInstance(obj, exp)


def raise_exception(*args, **kwargs):
    """Raises a ValueError exception.

    Args:
      *args: Callback arguments.
      *kwargs: Callback keyword arguments.
    Raises:
      ValueError: Unconditionally.
    """
    raise ValueError("Patched exception in parser")


class TestParserEntryTypes(unittest.TestCase):
    """Basic smoke test one entry of each kind."""

    @parser.parse_doc()
    def test_entry_transaction_one_string(self, entries, _, __):
        """
          2013-05-18 * "Nice dinner at Mermaid Inn"
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [data.Transaction])

    @parser.parse_doc()
    def test_entry_transaction_two_strings(self, entries, _, __):
        """
          2013-05-18 * "Mermaid Inn" "Nice dinner"
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [data.Transaction])

    @parser.parse_doc(expect_errors=True)
    def test_entry_transaction_three_strings(self, entries, errors, _):
        """
          2013-05-18 * "Mermaid Inn" "Nice dinner" "With Caroline"
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [])
        self.assertRegex(errors[0].message, "Too many strings")

    @parser.parse_doc()
    def test_entry_transaction_with_txn_keyword(self, entries, _, __):
        """
          2013-05-18 txn "Nice dinner at Mermaid Inn"
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [data.Transaction])

    @parser.parse_doc()
    def test_entry_balance(self, entries, _, __):
        """
          2013-05-18 balance Assets:US:BestBank:Checking  200 USD
          2013-05-18 balance Assets:US:BestBank:Checking  200 ~ 0.002 USD
        """
        check_list(self, entries, [data.Balance, data.Balance])

    @parser.parse_doc(expect_errors=True)
    def test_entry_balance_with_cost(self, entries, errors, __):
        """
          2013-05-18 balance Assets:Investments  10 MSFT {45.30 USD}
        """
        check_list(self, entries, [])
        check_list(self, errors, [parser.ParserSyntaxError])

    @parser.parse_doc()
    def test_entry_open_1(self, entries, _, __):
        """
          2013-05-18 open Assets:US:BestBank:Checking
        """
        check_list(self, entries, [data.Open])

    @parser.parse_doc()
    def test_entry_open_2(self, entries, _, __):
        """
          2013-05-18 open Assets:US:BestBank:Checking   USD
        """
        check_list(self, entries, [data.Open])

    @parser.parse_doc()
    def test_entry_open_3(self, entries, errors, __):
        """
          2013-05-18 open Assets:Cash   USD,CAD,EUR
        """
        check_list(self, entries, [data.Open])
        self.assertEqual(entries[0].booking, None)

    @parser.parse_doc()
    def test_entry_open_4(self, entries, errors, __):
        """
          2013-05-18 open Assets:US:Vanguard:VIIPX  VIIPX  "STRICT"
        """
        check_list(self, entries, [data.Open])
        self.assertEqual(entries[0].booking, data.Booking.STRICT)

    @parser.parse_doc()
    def test_entry_open_5(self, entries, errors, __):
        """
          2013-05-18 open Assets:US:Vanguard:VIIPX    "STRICT"
        """
        check_list(self, entries, [data.Open])
        self.assertEqual(entries[0].booking, data.Booking.STRICT)

    @parser.parse_doc()
    def test_entry_close(self, entries, _, __):
        """
          2013-05-18 close Assets:US:BestBank:Checking
        """
        check_list(self, entries, [data.Close])

    @parser.parse_doc()
    def test_entry_commodity(self, entries, _, __):
        """
          2013-05-18 commodity MSFT
        """
        check_list(self, entries, [data.Commodity])

    @parser.parse_doc()
    def test_entry_pad(self, entries, _, __):
        """
          2013-05-18 pad Assets:US:BestBank:Checking  Equity:Opening-Balances
        """
        check_list(self, entries, [data.Pad])

    @parser.parse_doc()
    def test_entry_event(self, entries, _, __):
        """
          2013-05-18 event "location" "New York, USA"

          ;; Test empty event.
          2013-05-18 event "location" ""
        """
        check_list(self, entries, [data.Event, data.Event])
        self.assertEqual("", entries[-1].description)

    @parser.parse_doc()
    def test_entry_query(self, entries, _, __):
        """
          2013-05-18 query "cash" "SELECT SUM(position) WHERE currency = 'USD'"
        """
        check_list(self, entries, [data.Query])

    @parser.parse_doc()
    def test_entry_note(self, entries, _, __):
        """
          2013-05-18 note Assets:US:BestBank:Checking  "Blah, di blah."
        """
        check_list(self, entries, [data.Note])

    @parser.parse_doc()
    def test_entry_document(self, entries, _, __):
        """
          2013-05-18 document Assets:US:BestBank:Checking "/Accounting/statement.pdf"
        """
        check_list(self, entries, [data.Document])

    @parser.parse_doc()
    def test_entry_price(self, entries, _, __):
        """
          2013-05-18 price USD   1.0290 CAD
        """
        check_list(self, entries, [data.Price])

    @parser.parse_doc()
    def test_entry_custom(self, entries, _, __):
        """
          2013-05-18 custom "budget" "weekly < 1000.00 USD" 2016-02-28 TRUE 43.03 USD 23
        """
        check_list(self, entries, [data.Custom])
        txns = [entry for entry in entries if isinstance(entry, data.Custom)]
        self.assertEqual([('weekly < 1000.00 USD', str),
                          (datetime.date(2016, 2, 28), datetime.date),
                          (True, bool),
                          (amount.from_string('43.03 USD'), amount.Amount),
                          (D('23'), Decimal)],
                         txns[0].values)


class TestParserComplete(unittest.TestCase):
    """Tests of completion of balance."""

    @parser.parse_doc()
    def test_entry_transaction_single_posting_at_zero(self, entries, errors, _):
        """
          2013-05-18 * "Nice dinner at Mermaid Inn"
            Expenses:Restaurant         0 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, 0)

    @parser.parse_doc()
    def test_entry_transaction_imbalance_from_single_posting(self, entries, errors, _):
        """
          2013-05-18 * "Nice dinner at Mermaid Inn"
            Expenses:Restaurant         100 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, 0)
        entry = entries[0]
        self.assertEqual(1, len(entry.postings))


class TestUglyBugs(unittest.TestCase):
    """Test all kinds of stupid sh*t that will inevitably occur in practice."""

    # pylint: disable=empty-docstring
    @parser.parse_doc()
    def test_empty_1(self, entries, errors, _):
        ""
        check_list(self, entries, [])
        check_list(self, errors, [])

    # pylint: disable=empty-docstring
    @parser.parse_doc()
    def test_empty_2(self, entries, errors, _):
        """

        """
        check_list(self, entries, [])
        check_list(self, errors, [])

    @parser.parse_doc()
    def test_comment(self, entries, errors, _):
        """
        ;; This is some comment.
        """
        check_list(self, entries, [])
        check_list(self, errors, [])

    def test_extra_whitespace_note(self):
        input_ = '\n2013-07-11 note Assets:Cash "test"\n\n  ;;\n'
        entries, errors, _ = parser.parse_string(input_)
        check_list(self, entries, [data.Note])
        check_list(self, errors, [])

    def test_extra_whitespace_transaction(self):
        input_ = '\n'.join([
            '2013-05-18 * "Nice dinner at Mermaid Inn"',
            '  Expenses:Restaurant         100 USD',
            '  Assets:US:Cash',
            '  ',
            ';; End of file',
        ])

        entries, errors, _ = parser.parse_string(input_)
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])

    def test_extra_whitespace_comment(self):
        input_ = '\n'.join([
            '2013-05-18 * "Nice dinner at Mermaid Inn"',
            '  Expenses:Restaurant         100 USD',
            '  Assets:US:Cash',
            '  ;;',
        ])
        entries, errors, _ = parser.parse_string(input_)
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])

    # pylint: disable=empty-docstring
    @parser.parse_doc()
    def test_indent_eof(self, entries, errors, _):
        "\t"
        check_list(self, entries, [])
        check_list(self, errors, [])

    @parser.parse_doc()
    def test_comment_eof(self, entries, errors, _):
        "; comment"
        check_list(self, entries, [])
        check_list(self, errors, [])

    @parser.parse_doc()
    def test_no_empty_lines(self, entries, errors, _):
        """
          2013-05-01 open Assets:Cash   USD,CAD,EUR
          2013-05-02 close Assets:US:BestBank:Checking
          2013-05-03 pad Assets:US:BestBank:Checking  Equity:Opening-Balances
          2013-05-04 event "location" "New York, USA"
          2013-05-05 * "Payee" "Narration"
            Assets:US:BestBank:Checking   100.00 USD
            Assets:Cash                  -100.00 USD
          2013-05-06 note Assets:US:BestBank:Checking  "Blah, di blah."
          2013-05-07 price USD   1.0290 CAD
        """
        self.assertEqual(7, len(entries))
        self.assertEqual([], errors)


class TestComment(unittest.TestCase):

    @parser.parse_doc()
    def test_comment_before_transaction(self, entries, errors, _):
        """
        ; Hi
        2015-06-07 *
          Assets:Cash   1 USD
          Assets:Cash   -1 USD
        """
        self.assertEqual(1, len(entries))
        self.assertEqual(2, len(entries[0].postings))
        self.assertEqual(0, len(errors))

    @parser.parse_doc()
    def test_comment_after_transaction(self, entries, errors, _):
        """
        2015-06-07 *
          Assets:Cash   1 USD
          Assets:Cash   -1 USD
        ; Hi
        """
        self.assertEqual(1, len(entries))
        self.assertEqual(2, len(entries[0].postings))
        self.assertEqual(0, len(errors))

    @parser.parse_doc()
    def test_comment_between_postings(self, entries, errors, _):
        """
        2015-06-07 *
          Assets:Cash   1 USD
          ; Hi
          Assets:Cash   -1 USD
        """
        self.assertEqual(1, len(entries))
        self.assertEqual(2, len(entries[0].postings))
        self.assertEqual(0, len(errors))

    @parser.parse_doc()
    def test_comment_after_posting(self, entries, errors, _):
        """
        2015-06-07 *
          Assets:Cash   1 USD    ; Hi
          Assets:Cash   -1 USD
        """
        self.assertEqual(1, len(entries))
        self.assertEqual(2, len(entries[0].postings))
        self.assertEqual(0, len(errors))

    @parser.parse_doc()
    def test_comment_after_transaction_start(self, entries, errors, _):
        """
        2015-06-07 *     ; Hi
          Assets:Cash   1 USD
          Assets:Cash   -1 USD
        """
        self.assertEqual(1, len(entries))
        self.assertEqual(2, len(entries[0].postings))
        self.assertEqual(0, len(errors))


class TestPushPopTag(unittest.TestCase):

    @parser.parse_doc(expect_errors=True)
    def test_tag_left_unclosed(self, entries, errors, _):
        """
          pushtag #trip-to-nowhere
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, 'Unbalanced pushed tag')

    @parser.parse_doc(expect_errors=True)
    def test_pop_invalid_tag(self, entries, errors, _):
        """
          poptag #trip-to-nowhere
        """
        self.assertTrue(errors)
        self.assertRegex(errors[0].message, 'absent tag')


class TestPushPopMeta(unittest.TestCase):

    @parser.parse_doc()
    def test_pushmeta_normal(self, entries, errors, _):
        """
          pushmeta location: "Lausanne, Switzerland"

          2015-06-07 * "Something"
            Assets:Something   1 USD
            Assets:Something  -1 USD

          popmeta location:
        """
        self.assertFalse(errors)
        self.assertTrue('location' in entries[0].meta)
        self.assertEqual("Lausanne, Switzerland", entries[0].meta['location'])

    @parser.parse_doc()
    def test_pushmeta_shadow(self, entries, errors, _):
        """
          pushmeta location: "Lausanne, Switzerland"

          2015-06-07 * "Something"
            location: "Paris, France"
            Assets:Something   1 USD
            Assets:Something  -1 USD

          popmeta location:
        """
        self.assertFalse(errors)
        self.assertTrue('location' in entries[0].meta)
        self.assertEqual("Paris, France", entries[0].meta['location'])

    @parser.parse_doc()
    def test_pushmeta_override(self, entries, errors, _):
        """
          pushmeta location: "Lausanne, Switzerland"

          2015-06-01 * "Something"
            Assets:Something   1 USD
            Assets:Something  -1 USD

          pushmeta location: "Paris, France"

          2015-06-02 * "Something"
            Assets:Something   1 USD
            Assets:Something  -1 USD

          popmeta location:
          popmeta location:
        """
        self.assertTrue('location' in entries[0].meta)
        self.assertEqual("Lausanne, Switzerland", entries[0].meta['location'])

        self.assertTrue('location' in entries[1].meta)
        self.assertEqual("Paris, France", entries[1].meta['location'])

    @parser.parse_doc(expect_errors=True)
    def test_pushmeta_invalid_pop(self, entries, errors, _):
        """
          popmeta location:
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message,
                                 "Attempting to pop absent metadata key")

    @parser.parse_doc(expect_errors=True)
    def test_pushmeta_forgotten(self, entries, errors, _):
        """
          pushmeta location: "Lausanne, Switzerland"
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message,
                                 "Unbalanced metadata key")


class TestMultipleLines(unittest.TestCase):

    @parser.parse_doc()
    def test_multiline_narration(self, entries, errors, _):
        """
          2014-07-11 * "Hello one line
          and yet another,
          and why not another!"
            Expenses:Restaurant         100 USD
            Assets:Cash                -100 USD
        """
        self.assertEqual(1, len(entries))
        self.assertFalse(errors)
        self.assertFalse(lexer.LexerError in map(type, errors))
        expected_narration = "Hello one line\nand yet another,\nand why not another!"
        self.assertEqual(expected_narration, entries[0].narration)


class TestSyntaxErrors(unittest.TestCase):
    """Test syntax errors that occur within the parser.
    One of our goals is to recover and report without ever
    bailing out with an exception."""

    @parser.parse_doc(expect_errors=True)
    def test_lexer_default_rule_1(self, entries, errors, _):
        """
          Account:*:Bla
        """
        self.assertEqual(entries, [])
        self.assertTrue(errors)
        self.assertTrue(lexer.LexerError in map(type, errors))

    @parser.parse_doc(expect_errors=True)
    def test_lexer_default_rule_2(self, entries, errors, _):
        """
          2013-05-18 * "Nice dinner at Mermaid Inn"
            Expenses:Resta(urant        100 USD
            Expenses:Tips                10 USD
            Assets:US:Cash             -110 USD

          2013-05-20 balance Assets:US:Cash  -110 USD
        """
        # This should fail to parse the "Expenses:Resta(urant" account name.

        # Check that we indeed read the 'check' entry that comes after the one
        # with the error.
        check_list(self, entries, [data.Balance])

        # Make sure at least one error is reported.
        self.assertEqual(1, len(errors))
        self.assertIsInstance(errors[0], lexer.LexerError)
        self.assertRegex(errors[0].message, 'Invalid token')

    @parser.parse_doc()
    def test_no_final_newline(self, entries, errors, _):
        """
          2014-11-02 *
            Assets:Something   1 USD
            Assets:Other      -1 USD"""
        self.assertFalse(errors)
        self.assertEqual(1, len(entries))
        self.assertEqual(2, len(entries[0].postings))


class TestLineNumbers(unittest.TestCase):
    """Check that the line numbers line up correctly."""

    @parser.parse_doc()
    def test_line_numbers(self, entries, errors, _):
        """
          2013-05-18 * "Nice dinner at Mermaid Inn"
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD

          2013-05-19 balance  Assets:US:Cash   -100 USD

          2013-05-20 note  Assets:US:Cash   "Something"

        """
        _, first_line = inspect.getsourcelines(
            TestLineNumbers.test_line_numbers.__wrapped__)
        first_line += 1

        self.assertEqual(2, entries[0].meta["lineno"] - first_line)
        self.assertEqual(6, entries[1].meta["lineno"] - first_line)
        self.assertEqual(8, entries[2].meta["lineno"] - first_line)


class TestParserOptions(unittest.TestCase):

    @parser.parse_doc()
    def test_option_single_value(self, entries, errors, options_map):
        """
          option "title" "Super Rich"

        """
        option = options_map['title']
        self.assertEqual(option, 'Super Rich')

    @parser.parse_doc()
    def test_option_list_value(self, entries, errors, options_map):
        """
          option "documents" "/path/docs/a"
          option "documents" "/path/docs/b"
          option "documents" "/path/docs/c"

        """
        documents = options_map['documents']
        self.assertEqual(['/path/docs/a',
                          '/path/docs/b',
                          '/path/docs/c'], documents)

    @parser.parse_doc(expect_errors=True)
    def test_invalid_option(self, entries, errors, options_map):
        """
          option "bladibla_invalid" "Some value"

        """
        check_list(self, errors, [parser.ParserError])
        self.assertFalse("bladibla_invalid" in options_map)

    @parser.parse_doc(expect_errors=True)
    def test_readonly_option(self, entries, errors, options_map):
        """
          option "filename" "gniagniagniagniagnia"

        """
        check_list(self, errors, [parser.ParserError])
        self.assertNotEqual("filename", "gniagniagniagniagnia")


class TestParserInclude(unittest.TestCase):

    def test_parse_nonexist(self):
        with self.assertRaises(OSError):
            parser.parse_file('/some/bullshit/filename.beancount')

    @test_utils.docfile
    def test_include_absolute(self, filename):
        """
          include "/some/absolute/filename.beancount"
        """
        entries, errors, options_map = parser.parse_file(filename)
        self.assertFalse(errors)
        self.assertEqual(['/some/absolute/filename.beancount'],
                         options_map['include'])

    @test_utils.docfile
    def test_include_relative(self, filename):
        """
          include "some/relative/filename.beancount"
        """
        entries, errors, options_map = parser.parse_file(filename)
        self.assertFalse(errors)
        self.assertEqual(['some/relative/filename.beancount'],
                         options_map['include'])

    def test_include_relative_from_string(self):
        input_string = """
          include "some/relative/filename.beancount"
        """
        entries, errors, options_map = parser.parse_string(input_string)
        self.assertFalse(errors)
        self.assertEqual(['some/relative/filename.beancount'],
                         options_map['include'])


class TestParserPlugin(unittest.TestCase):

    @parser.parse_doc()
    def test_plugin(self, entries, errors, options_map):
        """
          plugin "beancount.plugin.unrealized"
        """
        self.assertFalse(errors)
        self.assertEqual([('beancount.plugin.unrealized', None)],
                         options_map['plugin'])

    @parser.parse_doc()
    def test_plugin_with_config(self, entries, errors, options_map):
        """
          plugin "beancount.plugin.unrealized" "Unrealized"
        """
        self.assertFalse(errors)
        self.assertEqual([('beancount.plugin.unrealized', 'Unrealized')],
                         options_map['plugin'])

    @parser.parse_doc(expect_errors=True)
    def test_plugin_as_option(self, entries, errors, options_map):
        """
          option "plugin" "beancount.plugin.unrealized"
        """
        # Test that the very old method of plugin specification is disallowed.
        self.assertEqual(1, len(errors))
        self.assertEqual([], options_map['plugin'])


class TestDisplayContextOptions(unittest.TestCase):

    @parser.parse_doc()
    def test_render_commas_no(self, _, __, options_map):
        """
          option "render_commas" "0"
        """
        self.assertEqual(False, options_map['render_commas'])

    @parser.parse_doc()
    def test_render_commas_yes(self, _, __, options_map):
        """
          option "render_commas" "1"
        """
        self.assertEqual(True, options_map['render_commas'])

    @parser.parse_doc()
    def test_render_commas_yes2(self, _, __, options_map):
        """
          option "render_commas" "TRUE"
        """
        self.assertEqual(True, options_map['render_commas'])


class TestMiscOptions(unittest.TestCase):

    @parser.parse_doc(expect_errors=False)
    def test_plugin_processing_mode__default(self, _, __, options_map):
        """
        option "plugin_processing_mode" "default"
        """
        self.assertEqual("default", options_map['plugin_processing_mode'])

    @parser.parse_doc(expect_errors=False)
    def test_plugin_processing_mode__raw(self, _, __, options_map):
        """
        option "plugin_processing_mode" "raw"
        """
        self.assertEqual("raw", options_map['plugin_processing_mode'])

    @parser.parse_doc(expect_errors=True)
    def test_plugin_processing_mode__invalid(self, _, errors, options_map):
        """
        option "plugin_processing_mode" "invalid"
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, "Error for option")
        self.assertEqual("default", options_map['plugin_processing_mode'])


class TestToleranceOptions(unittest.TestCase):

    # pylint: disable=empty-docstring
    @parser.parse_doc()
    def test_tolerance_defaults(self, _, __, options_map):
        """
        """
        self.assertEqual({},
                         options_map['inferred_tolerance_default'])

    @parser.parse_doc()
    def test_inferred_tolerance_default(self, _, __, options_map):
        """
          option "inferred_tolerance_default" "*:0"
          option "inferred_tolerance_default" "USD:0.05"
          option "inferred_tolerance_default" "JPY:0.5"
        """
        self.assertEqual({"*": D("0"),
                          "USD": D("0.05"),
                          "JPY": D("0.5")},
                         options_map['inferred_tolerance_default'])


class TestDeprecatedOptions(unittest.TestCase):

    @parser.parse_doc(expect_errors=True)
    def test_deprecated_plugin(self, _, errors, __):
        """
          option "plugin" "beancount.plugins.module_name"
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, 'may not be set')

    @parser.parse_doc(expect_errors=True)
    def test_deprecated_option(self, _, errors, options_map):
        """
          option "allow_pipe_separator" "TRUE"
        """
        self.assertEqual(1, len(errors))
        self.assertEqual(True, options_map['allow_pipe_separator'])
        self.assertRegex(errors[0].message, "this will go away")


class TestParserLinks(unittest.TestCase):

    @parser.parse_doc()
    def test_links(self, entries, errors, _):
        """
          2013-05-18 * "Something something" ^38784734873
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD

        """
        check_list(self, entries, [data.Transaction])
        self.assertEqual(entries[0].links, set(['38784734873']))


class TestTransactions(unittest.TestCase):

    @parser.parse_doc()
    def test_simple_1(self, entries, errors, _):
        """
          2013-05-18 * "Nice dinner at Mermaid Inn"
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])
        self.assertEqual(None, entries[0].payee)
        self.assertEqual("Nice dinner at Mermaid Inn", entries[0].narration)

    @parser.parse_doc()
    def test_simple_2(self, entries, errors, _):
        """

          2013-05-18 * "Nice dinner at Mermaid Inn"
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD

          2013-05-20 * "Duane Reade" "Toothbrush"
            Expenses:BathroomSupplies         4 USD
            Assets:US:BestBank:Checking      -4 USD

        """
        check_list(self, entries, [data.Transaction, data.Transaction])
        check_list(self, errors, [])
        self.assertEqual(None, entries[0].payee)
        self.assertEqual("Nice dinner at Mermaid Inn", entries[0].narration)
        self.assertEqual("Duane Reade", entries[1].payee)
        self.assertEqual("Toothbrush", entries[1].narration)

    @parser.parse_doc()
    def test_empty_narration(self, entries, errors, _):
        """
          2013-05-18 * ""
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])
        self.assertEqual("", entries[0].narration)
        self.assertEqual(None, entries[0].payee)

    @parser.parse_doc()
    def test_no_narration(self, entries, errors, _):
        """
          2013-05-18 *
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])
        self.assertEqual("", entries[0].narration)
        self.assertEqual(None, entries[0].payee)

    @parser.parse_doc()
    def test_payee_no_narration(self, entries, errors, _):
        """
          2013-05-18 * "Mermaid Inn"
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])
        self.assertEqual(None, entries[0].payee)
        self.assertEqual("Mermaid Inn", entries[0].narration)

    @parser.parse_doc(expect_errors=True)
    def test_too_many_strings(self, entries, errors, _):
        """
          2013-05-18 * "A" "B" "C"
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [])
        check_list(self, errors, [parser.ParserError])

    @parser.parse_doc()
    def test_link_and_then_tag(self, entries, errors, _):
        """
          2014-04-20 * "Money from CC" ^610fa7f17e7a #trip
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])
        self.assertEqual("Money from CC", entries[0].narration)
        self.assertEqual(None, entries[0].payee)
        self.assertEqual(set(["610fa7f17e7a"]), entries[0].links)
        self.assertEqual(set(["trip"]), entries[0].tags)

    @parser.parse_doc(expect_errors=True)
    def test_tag_then_link(self, entries, errors, _):
        """
          2014-04-20 * #trip "Money from CC" ^610fa7f17e7a
            Expenses:Restaurant         100 USD
            Assets:US:Cash             -100 USD
        """
        check_list(self, entries, [])
        check_list(self, errors, [parser.ParserSyntaxError])

    @parser.parse_doc()
    def test_zero_prices(self, entries, errors, _):
        """
          2014-04-20 * "Like a conversion entry"
            Equity:Conversions         100 USD @ 0 XFER
            Equity:Conversions         101 CAD @ 0 XFER
            Equity:Conversions         102 AUD @ 0 XFER
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])

    @parser.parse_doc(expect_errors=False)
    def test_zero_units(self, entries, errors, _):
        """
          2014-04-20 * "Zero number of units"
            Assets:Investment         0 HOOL {500.00 USD}
            Assets:Cash               0 USD
        """
        check_list(self, entries, [data.Transaction])
        # Note: Zero amount is caught only at booking time.
        self.assertFalse(errors)

    @parser.parse_doc()
    def test_zero_costs(self, entries, errors, _):
        """
          2014-04-20 * "Like a conversion entry"
            Assets:Investment         10 HOOL {0 USD}
            Assets:Cash                0 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])

    @parser.parse_doc()
    def test_imbalance(self, entries, errors, _):
        """
          2014-04-20 * "Busted!"
            Assets:Checking         100 USD
            Assets:Checking         -99 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, errors, [])

    @parser.parse_doc()
    def test_no_postings(self, entries, errors, _):
        """
          2014-07-17 * "(JRN) INTRA-ACCOUNT TRANSFER" ^795422780
        """
        self.assertTrue(isinstance(entries[0].postings, list))

    @parser.parse_doc(expect_errors=True)
    def test_blank_line_not_allowed(self, entries, errors, _):
        """
          2014-04-20 * "Busted!"
            Assets:Checking         100 USD

            Assets:Checking         -99 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, entries[0].postings, [data.Posting])
        check_list(self, errors, [parser.ParserSyntaxError])

    def test_blank_line_with_spaces_not_allowed(self):
        input_ = '\n'.join([
            '2014-04-20 * "Busted!"',
            '  Assets:Checking         100 USD',
            '  ',  # This cuts off the transaction
            '  Assets:Checking         -99 USD'
        ])
        entries, errors, _ = parser.parse_string(input_)
        check_list(self, entries, [data.Transaction])
        check_list(self, entries[0].postings, [data.Posting])
        check_list(self, errors, [parser.ParserSyntaxError])

    @parser.parse_doc()
    def test_tags_after_first_line(self, entries, errors, _):
        """
          2014-04-20 * "Links and tags on subsequent lines" #basetag ^baselink
            #tag1 #tag2
            ^link1 #tag3
            #tag4 ^link2
            ^link3 ^link4
            #tag6
            ^link5
            Assets:Checking         100 USD
            Assets:Checking         -99 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, entries[0].postings, [data.Posting, data.Posting])
        check_list(self, errors, [])
        self.assertEqual({"basetag", "tag1", "tag2", "tag3", "tag4", "tag6"},
                         entries[0].tags)
        self.assertEqual({"baselink", "link1", "link2", "link3", "link4", "link5"},
                         entries[0].links)

    @parser.parse_doc(expect_errors=True)
    def test_tags_after_first_posting(self, entries, errors, _):
        """
          2014-04-20 * "Links and tags on subsequent lines" #basetag ^baselink
            Assets:Checking         100 USD
            #tag1 ^link1
            Assets:Checking         -99 USD
        """
        check_list(self, entries, [data.Transaction])
        check_list(self, entries[0].postings, [data.Posting, data.Posting])
        check_list(self, errors, [parser.ParserError])
        self.assertEqual({"basetag"}, entries[0].tags)
        self.assertEqual({"baselink"}, entries[0].links)


class TestParseLots(unittest.TestCase):

    maxDiff = None

    @parser.parse_doc()
    def test_cost_none(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL   45.23 USD
            Assets:Invest:Cash  -45.23 USD
        """
        self.assertFalse(parser.is_entry_incomplete(entries[0]))
        self.assertFalse(errors)
        posting = entries[0].postings[0]
        self.assertEqual(A('45.23 USD'), posting.units)
        self.assertEqual(None, posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_empty(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL   20 AAPL {}
            Assets:Invest:Cash  -20 AAPL
        """
        self.assertTrue(parser.is_entry_incomplete(entries[0]))
        self.assertFalse(errors)
        posting = entries[0].postings[0]
        self.assertEqual(A('20 AAPL'), posting.units)
        self.assertEqual(CostSpec(MISSING, None, MISSING, None, None, False), posting.cost)

    @parser.parse_doc()
    def test_cost_amount(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL      20 AAPL {45.23 USD}
            Assets:Invest:Cash  -90.46 USD
        """
        self.assertFalse(parser.is_entry_incomplete(entries[0]))
        self.assertFalse(errors)
        posting = entries[0].postings[0]
        self.assertEqual(A('20 AAPL'), posting.units)
        self.assertEqual(CostSpec(D('45.23'), None, 'USD', None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_date(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL   20 AAPL {2014-12-26}
            Assets:Invest:Cash  -20 AAPL
        """
        self.assertTrue(parser.is_entry_incomplete(entries[0]))
        self.assertFalse(errors)
        posting = entries[0].postings[0]
        self.assertEqual(A('20 AAPL'), posting.units)
        self.assertEqual(
            CostSpec(MISSING, None, MISSING, datetime.date(2014, 12, 26), None, False),
            posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_label(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL   20 AAPL {"d82d55a0dbe8"}
            Assets:Invest:Cash  -20 AAPL
        """
        self.assertTrue(parser.is_entry_incomplete(entries[0]))
        posting = entries[0].postings[0]
        self.assertEqual(A('20 AAPL'), posting.units)
        self.assertEqual(CostSpec(MISSING, None, MISSING, None, "d82d55a0dbe8", False),
                         posting.cost)

    @parser.parse_doc(allow_incomplete=True, expect_errors=True)
    def test_cost_merge(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL   20 AAPL {*}
            Assets:Invest:Cash  -20 AAPL
        """
        self.assertTrue(parser.is_entry_incomplete(entries[0]))
        posting = entries[0].postings[0]
        self.assertEqual(A('20 AAPL'), posting.units)
        self.assertEqual(CostSpec(MISSING, None, MISSING, None, None, True), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_two_components(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL    1 AAPL {45.23 USD, 2014-12-26}
            Assets:Invest:AAPL    1 AAPL {2014-12-26, 45.23 USD}
            Assets:Invest:AAPL    1 AAPL {45.23 USD, "d82d55a0dbe8"}
            Assets:Invest:AAPL    1 AAPL {"d82d55a0dbe8", 45.23 USD}
            Assets:Invest:AAPL    1 AAPL {2014-12-26, "d82d55a0dbe8"}
            Assets:Invest:AAPL    1 AAPL {"d82d55a0dbe8", 2014-12-26}
        """
        self.assertEqual(0, len(errors))
        self.assertTrue(parser.is_entry_incomplete(entries[0]))

    @parser.parse_doc()
    def test_cost_three_components(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL    1 AAPL {45.23 USD, 2014-12-26, "d82d55a0dbe8"}
            Assets:Invest:AAPL    1 AAPL {2014-12-26, 45.23 USD, "d82d55a0dbe8"}
            Assets:Invest:AAPL    1 AAPL {45.23 USD, "d82d55a0dbe8", 2014-12-26}
            Assets:Invest:AAPL    1 AAPL {2014-12-26, "d82d55a0dbe8", 45.23 USD}
            Assets:Invest:AAPL    1 AAPL {"d82d55a0dbe8", 45.23 USD, 2014-12-26}
            Assets:Invest:AAPL    1 AAPL {"d82d55a0dbe8", 2014-12-26, 45.23 USD}
        """
        self.assertEqual(0, len(errors))

    @parser.parse_doc(expect_errors=True)
    def test_cost_repeated(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL       1 AAPL {45.23 USD, 45.24 USD}
            Assets:Invest:Cash  -45.23 USD
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, "Duplicate cost")

    @parser.parse_doc(expect_errors=True)
    def test_cost_repeated_date(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL    1 AAPL {45.23 USD, 2014-12-26, 2014-12-27}
            Assets:Invest:Cash   -1 AAPL
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, "Duplicate date")

    @parser.parse_doc(expect_errors=True)
    def test_cost_repeated_label(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL       1 AAPL {"aaa", "bbb", 45.23 USD}
            Assets:Invest:Cash  -45.23 USD
        """
        self.assertEqual(1, len(errors))
        self.assertTrue(any(re.search("Duplicate label", error.message)
                            for error in errors))

    @parser.parse_doc(expect_errors=True, allow_incomplete=True)
    def test_cost_repeated_merge(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL       1 AAPL {*, *}
            Assets:Invest:Cash  -45.23 USD
        """
        self.assertTrue(parser.is_entry_incomplete(entries[0]))
        self.assertEqual(2, len(errors))
        self.assertTrue(any(re.search("Duplicate merge", error.message)
                            for error in errors))

    @parser.parse_doc()
    def test_cost_both_costs(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL       10 AAPL {45.23 # 9.95 USD}
            Assets:Invest:Cash  -110.36 USD
        """
        posting = entries[0].postings[0]
        self.assertEqual(CostSpec(D('45.23'), D('9.95'), 'USD', None, None, False),
                         posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_total_cost_only(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL      10 AAPL {# 9.95 USD}
            Assets:Invest:Cash  -19.90 USD
        """
        self.assertTrue(parser.is_entry_incomplete(entries[0]))
        posting = entries[0].postings[0]
        self.assertEqual(CostSpec(MISSING, D('9.95'), 'USD', None, None, False),
                         posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_total_empty_total(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL      20 AAPL {45.23 # USD}
            Assets:Invest:Cash  -45.23 USD
        """
        self.assertEqual(0, len(errors))
        self.assertTrue(parser.is_entry_incomplete(entries[0]))
        posting = entries[0].postings[0]
        self.assertEqual(A('20 AAPL'), posting.units)
        self.assertEqual(CostSpec(D('45.23'), MISSING, 'USD', None, None, False),
                         posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_total_just_currency(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL   20 AAPL {USD}
            Assets:Invest:AAPL   20 AAPL { # USD}
            Assets:Invest:Cash    0 USD
        """
        self.assertTrue(parser.is_entry_incomplete(entries[0]))

    @parser.parse_doc(expect_errors=True)
    def test_cost_with_slashes(self, entries, errors, _):
        """
          2014-01-01 *
            Assets:Invest:AAPL      1.1 AAPL {45.23 USD / 2015-07-16 / "blabla"}
            Assets:Invest:Cash   -45.23 USD
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, "unexpected SLASH")
        self.assertEqual(0, len(entries))


class TestCurrencies(unittest.TestCase):

    @parser.parse_doc()
    def test_parse_currencies(self, entries, errors, _):
        """
          2014-01-19 open Assets:Underscore    DJ_EURO
          2014-01-19 open Assets:Period        DJ.EURO
          2014-01-19 open Assets:Apostrophe    DJ'EURO
          2014-01-19 open Assets:Numbers       EURO123
        """
        self.assertFalse(errors)

    @parser.parse_doc(expect_errors=True)
    def test_different_cost_and_price_currency(self, entries, errors, _):
        """
          2018-03-21 * "Convert MR to KrisFlyer"
            Assets:Test                -100 MR {0.0075 USD} @ 1 KRISFLYER
            Assets:Krisflyer            100 KRISFLYER
        """


class TestTotalsAndSigns(unittest.TestCase):

    @parser.parse_doc(expect_errors=False)
    def test_zero_amount(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      0 MSFT {200.00 USD}
            Assets:Investments:Cash      0 USD
        """
        # Should produce no errors.

    @parser.parse_doc(expect_errors=False)
    def test_zero_cost(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      -10 MSFT {0.00 USD}
            Assets:Investments:Cash     0.00 USD
        """
        # Should produce no errors.

    @parser.parse_doc(expect_errors=False)
    def test_cost_negative(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      -10 MSFT {-200.00 USD}
            Assets:Investments:Cash  2000.00 USD
        """
        # Should produce no errors.
        # Note: This error is caught only at booking time.

    @parser.parse_doc()
    def test_total_cost(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      10 MSFT {{2,000 USD}}
            Assets:Investments:Cash  -20000 USD

          2013-05-18 * ""
            Assets:Investments:MSFT      10 MSFT {{2000 USD, 2014-02-25}}
            Assets:Investments:Cash  -20000 USD

          2013-06-01 * ""
            Assets:Investments:MSFT      -10 MSFT {{2,000 USD}}
            Assets:Investments:Cash    20000 USD
        """
        for entry in entries:
            posting = entry.postings[0]
            self.assertEqual(ZERO, posting.cost.number_per)
            self.assertEqual(D('2000'), posting.cost.number_total)
            self.assertEqual('USD', posting.cost.currency)
            self.assertEqual(None, posting.price)

    @parser.parse_doc(expect_errors=True)
    def test_total_cost__invalid(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      10 MSFT {{100 # 2,000 USD}}
            Assets:Investments:Cash  -20000 USD
        """
        posting = entries[0].postings[0]
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message,
                         'Per-unit cost may not be specified using total cost syntax')
        self.assertEqual(ZERO, posting.cost.number_per) # Note how this gets canceled.
        self.assertEqual(D('2000'), posting.cost.number_total)
        self.assertEqual('USD', posting.cost.currency)
        self.assertEqual(None, posting.price)

    @parser.parse_doc(expect_errors=False)
    def test_total_cost_negative(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      -10 MSFT {{-200.00 USD}}
            Assets:Investments:Cash   200.00 USD
        """
        # Should produce no errors.
        # Note: This error is caught only at booking time.

    @parser.parse_doc(expect_errors=True)
    def test_price_negative(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      -10 MSFT @ -200.00 USD
            Assets:Investments:Cash  2000.00 USD
        """
        self.assertRegex(errors[0].message, 'Negative.*allowed')

    @parser.parse_doc()
    def test_total_price_positive(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT        10 MSFT @@ 2000.00 USD
            Assets:Investments:Cash  -2000.00 USD
        """
        posting = entries[0].postings[0]
        self.assertEqual(amount.from_string('200 USD'), posting.price)
        self.assertEqual(None, posting.cost)

    @parser.parse_doc()
    def test_total_price_negative(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT       -10 MSFT @@ 2000.00 USD
            Assets:Investments:Cash  20000.00 USD
        """
        posting = entries[0].postings[0]
        self.assertEqual(amount.from_string('200 USD'), posting.price)
        self.assertEqual(None, posting.cost)

    @parser.parse_doc(expect_errors=True)
    def test_total_price_inverted(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT         10 MSFT @@ -2000.00 USD
            Assets:Investments:Cash   20000.00 USD
        """
        self.assertRegex(errors[0].message, 'Negative.*allowed')


class TestBalance(unittest.TestCase):

    @parser.parse_doc()
    def test_total_price(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      10 MSFT @@ 2000 USD
            Assets:Investments:Cash  -20000 USD
        """
        posting = entries[0].postings[0]
        self.assertEqual(amount.from_string('200 USD'), posting.price)
        self.assertEqual(None, posting.cost)

    @parser.parse_doc()
    def test_total_cost(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      10 MSFT {{2,000 USD}}
            Assets:Investments:Cash  -20000 USD

          2013-05-18 * ""
            Assets:Investments:MSFT      10 MSFT {{2000 USD, 2014-02-25}}
            Assets:Investments:Cash  -20000 USD
        """
        for entry in entries:
            posting = entry.postings[0]
            self.assertEqual(ZERO, posting.cost.number_per)
            self.assertEqual(D('2000'), posting.cost.number_total)
            self.assertEqual('USD', posting.cost.currency)
            self.assertEqual(None, posting.price)


class TestMetaData(unittest.TestCase):

    @staticmethod
    def strip_meta(meta):
        """Removes the filename, lineno from the postings metadata."""
        copy = meta.copy()
        copy.pop('filename', None)
        copy.pop('lineno', None)
        return copy

    @parser.parse_doc()
    def test_metadata_transaction__begin(self, entries, errors, _):
        """
          2013-05-18 * ""
            test: "Something"
            Assets:Investments:MSFT      10 MSFT @@ 2000 USD
            Assets:Investments:Cash  -20000 USD
        """
        self.assertEqual(1, len(entries))
        self.assertEqual('Something', entries[0].meta['test'])

    @parser.parse_doc()
    def test_metadata_transaction__middle(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      10 MSFT @@ 2000 USD
            test: "Something"
            Assets:Investments:Cash  -20000 USD
        """
        self.assertEqual(1, len(entries))
        self.assertEqual({'test': 'Something'},
                         self.strip_meta(entries[0].postings[0].meta))

    @parser.parse_doc()
    def test_metadata_transaction__end(self, entries, errors, _):
        """
          2013-05-18 * ""
            Assets:Investments:MSFT      10 MSFT @@ 2000 USD
            Assets:Investments:Cash  -20000 USD
            test: "Something"
        """
        self.assertEqual(1, len(entries))
        self.assertEqual({'test': 'Something'},
                         self.strip_meta(entries[0].postings[1].meta))

    @parser.parse_doc()
    def test_metadata_transaction__many(self, entries, errors, _):
        """
          2013-05-18 * ""
            test1: "Something"
            Assets:Investments:MSFT      10 MSFT @@ 2000 USD
            test2: "has"
            test3: "to"
            Assets:Investments:Cash  -20000 USD
            test4: "come"
            test5: "from"
            test6: "this"
        """
        self.assertEqual(1, len(entries))
        self.assertEqual('Something', entries[0].meta['test1'])
        self.assertEqual({'test2': 'has', 'test3': 'to'},
                         self.strip_meta(entries[0].postings[0].meta))
        self.assertEqual({'test4': 'come', 'test5': 'from', 'test6': 'this'},
                         self.strip_meta(entries[0].postings[1].meta))

    @parser.parse_doc()
    def test_metadata_transaction__indented(self, entries, errors, _):
        """
          2013-05-18 * ""
              test1: "Something"
            Assets:Investments:MSFT      10 MSFT @@ 2000 USD
              test2: "has"
              test3: "to"
            Assets:Investments:Cash  -20000 USD
              test4: "come"
              test5: "from"
              test6: "this"
        """
        self.assertEqual(1, len(entries))
        self.assertEqual('Something', entries[0].meta['test1'])
        self.assertEqual({'test2': 'has', 'test3': 'to'},
                         self.strip_meta(entries[0].postings[0].meta))
        self.assertEqual({'test4': 'come', 'test5': 'from', 'test6': 'this'},
                         self.strip_meta(entries[0].postings[1].meta))

    @parser.parse_doc(expect_errors=True)
    def test_metadata_transaction__repeated(self, entries, errors, _):
        """
          2013-05-18 * ""
            test: "Bananas"
            test: "Apples"
            test: "Oranges"
            Assets:Investments   100 USD
              test: "Bananas"
              test: "Apples"
            Income:Investments  -100 USD
        """
        self.assertEqual(1, len(entries))
        self.assertEqual('Bananas', entries[0].meta['test'])
        self.assertEqual({'test': 'Bananas'},
                         self.strip_meta(entries[0].postings[0].meta))
        self.assertEqual(3, len(errors))
        self.assertTrue(all(re.search('Duplicate.*metadata field', error.message)
                            for error in errors))

    @parser.parse_doc()
    def test_metadata_empty(self, entries, errors, _):
        """
          2013-05-18 * "blabla"
            oranges:
            bananas:

          2013-05-19 open Assets:Something
            apples:
        """
        self.assertFalse(errors)
        self.assertEqual(2, len(entries))
        self.assertEqual({'oranges', 'bananas', 'filename', 'lineno'},
                         entries[0].meta.keys())
        self.assertEqual(None, entries[0].meta['oranges'])
        self.assertEqual(None, entries[0].meta['bananas'])
        self.assertEqual(entries[1].meta['apples'], None)

    @parser.parse_doc()
    def test_metadata_other(self, entries, errors, _):
        """
          2013-01-01 open Equity:Other

          2013-01-01 open Assets:Investments
            test1: "Something"
            test2: "Something"

          2014-01-01 close Assets:Investments
            test1: "Something"

          2013-01-10 note Assets:Investments "Bla"
            test1: "Something"

          2013-01-31 pad Assets:Investments Equity:Other
            test1: "Something"

          2013-02-01 balance Assets:Investments  111.00 USD
            test1: "Something"

          2013-03-01 event "location" "Nowhere"
            test1: "Something"

          2013-03-01 document Assets:Investments "/path/to/something.pdf"
            test1: "Something"

          2013-03-01 price  HOOL  500 USD
            test1: "Something"
        """
        self.assertEqual(9, len(entries))

    @parser.parse_doc()
    def test_metadata_data_types(self, entries, errors, _):
        """
          2013-05-18 * ""
            string: "Something"
            account: Assets:Investments:Cash
            date: 2012-01-01
            currency: HOOL
            tag: #trip-florida
            number: 345.67
            amount: 345.67 USD
            boolt: TRUE
            boolf: FALSE
        """
        self.assertEqual(1, len(entries))
        self.assertTrue('filename' in entries[0].meta)
        self.assertTrue('lineno' in entries[0].meta)
        del entries[0].meta['filename']
        del entries[0].meta['lineno']
        self.assertEqual({
            'string': 'Something',
            'account': 'Assets:Investments:Cash',
            'date': datetime.date(2012, 1, 1),
            'currency': 'HOOL',
            'tag': 'trip-florida',
            'number': D('345.67'),
            'amount': A('345.67 USD'),
            'boolt': True,
            'boolf': False,
            }, entries[0].meta)

    @parser.parse_doc()
    def test_metadata_key_syntax(self, entries, errors, _):
        """
          2013-05-18 * ""
            nameoncard: "Jim"
            nameOnCard: "Joe"
            name-on-card: "Bob"
            name_on_card: "John"
        """
        self.assertEqual(1, len(entries))
        self.assertLessEqual(set('nameoncard nameOnCard name-on-card name_on_card'.split()),
                             set(entries[0].meta.keys()))


class TestArithmetic(unittest.TestCase):

    maxDiff = None

    @parser.parse_doc()
    def test_number_expr__add(self, entries, errors, _):
        """
          2013-05-18 * "Test"
            Assets:Something    12 + 3 USD
            Assets:Something   7.5 + 3.1 USD
        """
        self.assertEqual(1, len(entries))
        postings = entries[0].postings
        self.assertEqual(D('15'), postings[0].units.number)
        self.assertEqual(D('10.6'), postings[1].units.number)

    @parser.parse_doc()
    def test_number_expr__subtract(self, entries, errors, _):
        """
          2013-05-18 * "Test"
            Assets:Something    12 - 3 USD
            Assets:Something   7.5 - 3.1 USD
        """
        self.assertEqual(1, len(entries))
        postings = entries[0].postings
        self.assertEqual(D('9'), postings[0].units.number)
        self.assertEqual(D('4.4'), postings[1].units.number)

    @parser.parse_doc()
    def test_number_expr__multiply(self, entries, errors, _):
        """
          2013-05-18 * "Test"
            Assets:Something    12 * 3 USD
            Assets:Something   7.5 * 3.1 USD
        """
        self.assertEqual(1, len(entries))
        postings = entries[0].postings
        self.assertEqual(D('36'), postings[0].units.number)
        self.assertEqual(D('23.25'), postings[1].units.number)

    @parser.parse_doc()
    def test_number_expr__divide(self, entries, errors, _):
        """
          2013-05-18 * "Test"
            Assets:Something    12 / 3 USD
            Assets:Something   7.5 / 3 USD
        """
        self.assertEqual(1, len(entries))
        postings = entries[0].postings
        self.assertEqual(D('4'), postings[0].units.number)
        self.assertEqual(D('2.5'), postings[1].units.number)

    @parser.parse_doc()
    def test_number_expr__negative(self, entries, errors, _):
        """
          2013-05-18 * "Test"
            Assets:Something    -12 USD
            Assets:Something   -7.5 USD
            Assets:Something   - 7.5 USD
        """
        self.assertEqual(1, len(entries))
        postings = entries[0].postings
        self.assertEqual(D('-12'), postings[0].units.number)
        self.assertEqual(D('-7.5'), postings[1].units.number)
        self.assertEqual(D('-7.5'), postings[2].units.number)

    @parser.parse_doc()
    def test_number_expr__positive(self, entries, errors, _):
        """
          2013-05-18 * "Test"
            Assets:Something    +12 USD
            Assets:Something   -7.5 USD
        """
        self.assertEqual(1, len(entries))
        postings = entries[0].postings
        self.assertEqual(D('12'), postings[0].units.number)

    @parser.parse_doc()
    def test_number_expr__precedence(self, entries, errors, _):
        """
          2013-05-18 * "Test"
            Assets:Something   2 * 3 + 4 USD
            Assets:Something   2 + 3 * 4 USD
            Assets:Something   2 + -3 * 4 USD
            Assets:Something   (2 + -3) * 4 USD
        """
        self.assertEqual(1, len(entries))
        self.assertListEqual(
            [D('10'), D('14'), D('-10'), D('-4')],
            [posting.units.number for posting in entries[0].postings])

    @parser.parse_doc()
    def test_number_expr__groups(self, entries, errors, _):
        """
          2013-05-18 * "Test"
            Assets:Something   (2 + -3) * 4 USD
            Assets:Something   2 * (2 + -3) USD
        """
        self.assertEqual(1, len(entries))
        self.assertListEqual(
            [D('-4'), D('-2')],
            [posting.units.number
             for posting in entries[0].postings])

    @parser.parse_doc()
    def test_number_expr__different_places(self, entries, errors, _):
        """
          2013-05-18 * "Test"
            Assets:Something   -(3 * 4) HOOL {120.01 * 2.1 USD} @ 134.02 * 2.1 USD
            Assets:Something   1000000 USD ;; No balance checks.
          2014-01-01 balance Assets:Something  3 * 4 * 120.01 * 2.1  USD
            number: -(5662.23 + 22.3)
        """
        self.assertFalse(errors)
        self.assertEqual(2, len(entries))
        self.assertEqual(D('-12'), entries[0].postings[0].units.number)
        self.assertEqual(D('252.021'),
                         entries[0].postings[0].cost.number_per)
        self.assertEqual(None,
                         entries[0].postings[0].cost.number_total)
        self.assertEqual(D('281.442'), entries[0].postings[0].price.number)
        self.assertEqual(D('3024.252'), entries[1].amount.number)
        self.assertEqual(D('-5684.53'), entries[1].meta['number'])


class TestLexerAndParserErrors(cmptest.TestCase):
    """There are a number of different paths where errors may occur. This test case
    intends to exercise them all. This docstring explains how it all works (it's
    a bit complicated).

    Expectations: The parser.parse_*() functions never raise exceptions, they
    always produce a list of error messages on the builder objects. This is by
    design: we have a single way to produce and report errors.

    The lexer may fail in two different ways:

    * Invalid Token: yylex() is invoked by the parser... the lexer rules are
      processed and all fail and the default rule of the lexer gets called at
      {bf253a29a820}. The lexer immediately switches to its INVALID subparser to
      chomp the rest of the invalid token's unrecognized characters until it
      hits some whitespace (see {bba169a1d35a}). A new LexerError is created and
      accumulated by calling the build_lexer_error() method on the builder
      object using this text (see {0e31aeca3363}). The lexer then returns a
      LEX_ERROR token to the parser, which fails as below.

    * Lexer Builder Exception: The lexer recognizes a valid token and invokes a
      callback on the builder using BUILD_LEX(). However, an exception is raised
      in that Python code ({3cfb2739349a}). The exception's error text is saved
      and this is used to build an error on the lexer builder at
      build_lexer_error().

    For both the errors above, when the parser receives the LEX_ERROR token, it
    is an unexpected token (because none of the rules handle it), so it calls
    yyerror() {ca6aab8b9748}. Because the lexer has already registered an error
    in the error list, in yyerror() we simply ignore it and do nothing (see
    {ca6aab8b9748}).

    Error recovery then proceeds by successive reductions of the "error" grammar
    rules which discards all tokens until a valid rule can be reduced again
    ({3d95e55b654e}). Note that Bison issues a single call to yyerror() and
    keeps reducing invalid "error" rules silently until another one succeeds. We
    ignore the directive and restart parsing from that point on. If the error
    occurred on a posting in progress of being parsed, because the "error" rule
    is not a valid posting of a transaction, that transaction will not be
    produced and thus is ignore, which is what we want.

    The parser may itself also encounter two similar types of errors:

    * Syntax Error: There is an error in the grammar of the input. yyparse() is
      called automatically by the generated code and we call
      build_grammar_error() to register the error. Error recovery proceeds
      similarly to what was described previously.

    * Grammar Builder Exception: A grammar rule is reduced successfully, a
      builder method is invoked and raises a Python exception. A macro in the
      code that invokes this method is used to catch this error and calls
      build_grammar_error_from_exception() to register an error and makes the
      parser issue an error with YYERROR (see {05bb0fb60e86}).

    We never call YYABORT anywhere so the yyparse() function should never return
    anything else than 0. If it does, we translate that into a Python
    RuntimeError exception, or a MemoryError exception (if yyparse() ran out of
    memory), see {459018e2905c}.
    """

    @parser.parse_doc(expect_errors=True)
    def test_lexer_invalid_token(self, entries, errors, _):
        """
          2000-01-01 open ) USD
        """
        self.assertEqual(0, len(entries))
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, r"syntax error, unexpected RPAREN")

    @parser.parse_doc(expect_errors=True)
    def test_lexer_invalid_token__recovery(self, entries, errors, _):
        """
          2000-01-01 open ) USD
          2000-01-02 open Assets:Something
        """
        self.assertEqualEntries("""
          2000-01-02 open Assets:Something
        """, entries)
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, r"syntax error, unexpected RPAREN")

    @parser.parse_doc(expect_errors=True)
    def test_lexer_exception(self, entries, errors, _):
        """
          2000-13-32 open Assets:Something
        """
        self.assertEqual(0, len(entries))
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, 'month must be in 1..12')

    @parser.parse_doc(expect_errors=True)
    def test_lexer_exception__recovery(self, entries, errors, _):
        """
          2000-13-32 open Assets:Something
          2000-01-02 open Assets:Working
        """
        self.assertEqualEntries("""
          2000-01-02 open Assets:Working
        """, entries)
        self.assertEqual(1, len(entries))
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, 'month must be in 1..12')

    def test_lexer_errors_in_postings(self):
        txn_strings = textwrap.dedent("""

          2000-01-02 *
            Assets:Working  `
            Assets:Working  11.11 USD
            Assets:Working  22.22 USD

          2000-01-02 *
            Assets:Working  11.11 USD
            Assets:Working  `
            Assets:Working  22.22 USD

          2000-01-02 *
            Assets:Working  11.11 USD
            Assets:Working  22.22 USD
            Assets:Working  `

          2000-01-02 *
            Assets:Working  )
            Assets:Working  11.11 USD
            Assets:Working  22.22 USD

          2000-01-02 *
            Assets:Working  11.11 USD
            Assets:Working  )
            Assets:Working  22.22 USD

          2000-01-02 *
            Assets:Working  11.11 USD
            Assets:Working  22.22 USD
            Assets:Working  )

        """).strip().split('\n\n')
        self.assertEqual(6, len(txn_strings))
        for txn_string in txn_strings:
            input_string = txn_string + textwrap.dedent("""
              2000-01-02 open Assets:Working
            """)
            entries, errors, _ = parser.parse_string(input_string)

            # Check that the transaction is not produced.
            self.assertEqualEntries("""
              2000-01-02 open Assets:Working
            """, entries)
            self.assertEqual(1, len(entries))
            self.assertEqual(1, len(errors))
            self.assertRegex(errors[0].message,
                                     '(Invalid token|unexpected RPAREN)')

    @parser.parse_doc(expect_errors=True)
    def test_grammar_syntax_error(self, entries, errors, _):
        """
          2000-01-01 open open
        """
        self.assertEqual(0, len(entries))
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, r"syntax error")

    @parser.parse_doc(expect_errors=True)
    def test_grammar_syntax_error__recovery(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          2000-01-02 open open
          2000-01-03 open Assets:After
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, r"syntax error")
        self.assertEqualEntries("""
          2000-01-01 open Assets:Before
          2000-01-03 open Assets:After
        """, entries)

    @parser.parse_doc(expect_errors=True)
    def test_grammar_syntax_error__recovery2(self, entries, errors, _):
        """
          2000-01-01 open open
          2000-01-02 open Assets:Something
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, r"syntax error")
        self.assertEqual(1, len(entries))
        self.assertEqualEntries("""
          2000-01-02 open Assets:Something
        """, entries)

    @parser.parse_doc(expect_errors=True)
    def test_grammar_syntax_error__multiple(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          2000-01-02 open open
          2000-01-03 open Assets:After
          2000-01-04 close close
          2000-01-05 close Assets:After
        """
        self.assertEqual(2, len(errors))
        for error in errors:
            self.assertRegex(error.message, r"syntax error")
        self.assertEqual(3, len(entries))
        self.assertEqualEntries("""
          2000-01-01 open Assets:Before
          2000-01-03 open Assets:After
          2000-01-05 close Assets:After
        """, entries)

    def check_entries_errors(self, entries, errors):
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, 'Patched exception')
        self.assertEqual(2, len(entries))

    @mock.patch('beancount.parser.grammar.Builder.pushtag', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__pushtag(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          pushtag #sometag
          2010-01-01 close Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.poptag', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__poptag(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          poptag #sometag
          2010-01-01 close Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.option', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__option(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          option "operating_currency" "CAD"
          2010-01-01 close Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.include', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__include(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          include "answer.beancount"
          2010-01-01 close Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.plugin', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__plugin(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          plugin "answer.beancount"
          2010-01-01 close Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.amount', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__amount(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          2001-02-02 balance Assets:Before    23.00 USD
          2010-01-01 close Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.compound_amount', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__compound_amount(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          2001-02-02 *
            Assets:Before   10.00 HOOL {100.00 # 9.95 USD}
            Assets:After   -100.00 USD
          2010-01-01 close Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.cost_spec', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__lot_cost_date(self, entries, errors, _):
        """
          2000-01-01 open Assets:Before
          2001-02-02 *
            Assets:Before   10.00 HOOL {100.00 USD}
            Assets:After   -100.00 USD
          2010-01-01 close Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.open', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__open(self, entries, errors, _):
        """
          2010-01-01 balance Assets:Before  1 USD
          2000-01-01 open Assets:Before
          2010-01-01 close Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.close', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__close(self, entries, errors, _):
        """
          2010-01-01 balance Assets:Before  1 USD
          2010-01-01 close Assets:Before
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.commodity', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__commodity(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 commodity USD
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.pad', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__pad(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 pad Assets:Before Assets:After
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.balance', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__balance(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 balance Assets:Before 100 USD
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.event', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__event(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 event "location" "New York, NY"
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.price', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__price(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 price HOOL 20 USD
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.note', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__note(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 note Assets:Before "Something something"
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.document', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__document(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 document Assets:Before "/path/to/document.png"
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.key_value', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__key_value(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 commodity HOOL
            key: "Value"
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.posting', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__posting(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 *
            Assets:Before   100.00 USD
            Assets:After   -100.00 USD
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.tag_link_new', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__tag_link_new(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 * "Payee" "Narration"
            Assets:Before   100.00 USD
            Assets:After   -100.00 USD
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.tag_link_TAG', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__tag_link_TAG(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 * "Payee" "Narration" #sometag
            Assets:Before   100.00 USD
            Assets:After   -100.00 USD
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @mock.patch('beancount.parser.grammar.Builder.tag_link_LINK', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__tag_link_LINK(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 * "Payee" "Narration" ^somelink
            Assets:Before   100.00 USD
            Assets:After   -100.00 USD
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)

    @parser.parse_doc()
    def test_grammar_exceptions__tag_link_PIPE(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 * "Payee" "Narration"
            Assets:Before   100.00 USD
            Assets:After   -100.00 USD
          2000-01-01 open Assets:Before
        """

    @mock.patch('beancount.parser.grammar.Builder.transaction', raise_exception)
    @parser.parse_doc(expect_errors=True)
    def test_grammar_exceptions__transaction(self, entries, errors, _):
        """
          2010-01-01 close Assets:Before
          2010-01-01 * "Payee" "Narration"
            Assets:Before   100.00 USD
            Assets:After   -100.00 USD
          2000-01-01 open Assets:Before
        """
        self.check_entries_errors(entries, errors)


class TestIncompleteInputs(cmptest.TestCase):

    #
    # Units
    #

    @parser.parse_doc(allow_incomplete=True)
    def test_units_full(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     100.00 USD
            Assets:Account2    -100.00 USD
        """
        self.assertEqual(A('-100 USD'), entries[-1].postings[-1].units)

    @parser.parse_doc(allow_incomplete=True)
    def test_units_missing(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     100.00 USD
            Assets:Account2
        """
        self.assertEqual(MISSING, entries[-1].postings[-1].units)
        self.assertEqual(None, entries[-1].postings[-1].cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_units_missing_number(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     100.00 USD
            Assets:Account2            USD
        """
        units = entries[-1].postings[-1].units
        self.assertEqual(Amount(MISSING, 'USD'), units)

    @parser.parse_doc(allow_incomplete=True)
    def test_units_missing_currency(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     100.00 USD
            Assets:Account2    -100.00
        """
        units = entries[-1].postings[-1].units
        self.assertEqual(Amount(D('-100.00'), MISSING), units)

    @parser.parse_doc(allow_incomplete=True)
    def test_units_missing_with_cost(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     {300.00 USD}
            Assets:Account2    -600.00 USD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(Amount(MISSING, MISSING), posting.units)
        self.assertEqual(CostSpec(D('300'), None, 'USD', None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_units_missing_number_with_cost(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1            HOOL {300.00 USD}
            Assets:Account2    -600.00 USD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(Amount(MISSING, 'HOOL'), posting.units)
        self.assertEqual(CostSpec(D('300'), None, 'USD', None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_units_missing_currency_with_cost(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1      10        {300.00 USD}
            Assets:Account2    -600.00 USD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(Amount(D('10'), MISSING), posting.units)
        self.assertEqual(CostSpec(D('300'), None, 'USD', None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_units_missing_with_price(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account2                @ 1.2 USD
            Assets:Account1     100.00 USD @
        """
        posting = entries[-1].postings[0]
        self.assertEqual(Amount(MISSING, MISSING), posting.units)
        self.assertEqual(Amount(D('1.2'), 'USD'), posting.price)

    @parser.parse_doc(allow_incomplete=True)
    def test_units_missing_number_with_price(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account2            CAD @ 1.2 USD
            Assets:Account1     100.00 USD @
        """
        posting = entries[-1].postings[0]
        self.assertEqual(Amount(MISSING, 'CAD'), posting.units)
        self.assertEqual(Amount(D('1.2'), 'USD'), posting.price)

    @parser.parse_doc(allow_incomplete=True)
    def test_units_missing_currency_with_price(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account2     120.00     @ 1.2 USD
            Assets:Account1     100.00 USD @
        """
        posting = entries[-1].postings[0]
        self.assertEqual(Amount(D('120.00'), MISSING), posting.units)
        self.assertEqual(Amount(D('1.2'), 'USD'), posting.price)

    #
    # Price
    #

    @parser.parse_doc(allow_incomplete=True)
    def test_price_none(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     100.00 USD
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(None, posting.cost)
        self.assertEqual(None, posting.price)

    @parser.parse_doc(allow_incomplete=True)
    def test_price_missing(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     100.00 USD @
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(A('100.00 USD'), posting.units)
        self.assertIsInstance(posting.price, Amount)
        self.assertEqual(Amount(MISSING, MISSING), posting.price)

    @parser.parse_doc(allow_incomplete=True)
    def test_price_missing_number(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     100.00 USD @ CAD
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertIsInstance(posting.price, Amount)
        self.assertEqual(Amount(MISSING, 'CAD'), posting.price)

    @parser.parse_doc(allow_incomplete=True)
    def test_price_missing_currency(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     100.00 USD @ 1.2
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertIsInstance(posting.price, Amount)
        self.assertEqual(Amount(D('1.2'), MISSING), posting.price)


    #
    # Cost
    #

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_full(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {150 # 5 USD}
            Assets:Account2     120.00 USD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(D('150'), D('5'), 'USD', None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_missing_number_per(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {# 5 USD}
            Assets:Account2     120.00 USD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(MISSING, D('5'), 'USD', None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_missing_number_total(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {150 # USD}
            Assets:Account2     120.00 USD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(D('150'), MISSING, 'USD', None, None, False),
                         posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_no_number_total(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {150 USD}
            Assets:Account2     120.00 USD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(D('150'), None, 'USD', None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_missing_numbers(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {USD}
            Assets:Account2     120.00 USD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(MISSING, None, 'USD', None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_missing_currency(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {150}
            Assets:Account2     120.00 USD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(D('150'), None, MISSING, None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_empty(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {}
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(MISSING, None, MISSING, None, None, False), posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_empty_with_other(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {2015-09-21, "blablabla"}
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(MISSING, None, MISSING,
                                  datetime.date(2015, 9, 21), "blablabla", False),
                         posting.cost)

    @parser.parse_doc(allow_incomplete=True)
    def test_cost_missing_basis(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {2015-09-21, "blablabla"}
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(MISSING, None, MISSING,
                                  datetime.date(2015, 9, 21), "blablabla", False),
                         posting.cost)

    @parser.parse_doc(allow_incomplete=True, expect_errors=True)
    def test_cost_average(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {*}
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(MISSING, None, MISSING, None, None, True),
                         posting.cost)

    @parser.parse_doc(allow_incomplete=True, expect_errors=True)
    def test_cost_average_missing_basis(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {*, 2015-09-21, "blablabla"}
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(MISSING, None, MISSING,
                                  datetime.date(2015, 9, 21), "blablabla", True),
                         posting.cost)

    @parser.parse_doc(allow_incomplete=True, expect_errors=True)
    def test_cost_average_with_other(self, entries, _, options_map):
        """
          2010-05-28 *
            Assets:Account1     2 HOOL {*, 100.00 CAD, 2015-09-21, "blablabla"}
            Assets:Account2     120.00 CAD
        """
        posting = entries[-1].postings[0]
        self.assertEqual(CostSpec(D("100.00"), None, "CAD",
                                  datetime.date(2015, 9, 21), "blablabla", True),
                         posting.cost)


class TestMisc(cmptest.TestCase):

    @parser.parse_doc(expect_errors=False)
    def test_comment_in_postings(self, entries, errors, options_map):
        """
          2017-06-27 * "Bitcoin network fee"
            ; Account: Pocket money
            Expenses:Crypto:NetworkFees           0.00082487 BTC
            Assets:Crypto:Bitcoin                -0.00082487 BTC
        """
        self.assertEqual(0, len(errors))

    # FIXME(blais): Ideally this unindented should generate an error.
    # It would be nicer if only indented comments would be allowed.
    @parser.parse_doc(expect_errors=True)
    def test_comment_in_postings_invalid(self, entries, errors, options_map):
        """
          2017-06-27 * "Bitcoin network fee"
            Expenses:Crypto:NetworkFees           0.00082487 BTC
          ; Account: Pocket money
            Assets:Crypto:Bitcoin                -0.00082487 BTC
        """
        self.assertEqual(1, len(errors))
        self.assertRegex(errors[0].message, "unexpected ACCOUNT")


class TestDocument(unittest.TestCase):

    @parser.parse_doc()
    def test_document_no_tags_links(self, entries, _, __):
        """
          2013-05-18 document Assets:US:BestBank:Checking "/Accounting/statement.pdf"
        """
        check_list(self, entries, [data.Document])

    @parser.parse_doc()
    def test_document_tags(self, entries, _, __):
        """
          pushtag #something
          2013-05-18 document Assets:US:BestBank:Checking "/Accounting/statement.pdf" #else
          poptag #something
        """
        check_list(self, entries, [data.Document])
        self.assertEqual({'something', 'else'}, entries[0].tags)

    @parser.parse_doc()
    def test_document_links(self, entries, _, __):
        """
          2013-05-18 document Assets:US:BestBank:Checking "/statement.pdf" ^something
        """
        check_list(self, entries, [data.Document])
        self.assertEqual({'something'}, entries[0].links)


class TestMethodsSignature(unittest.TestCase):

    def test_signatures(self):
        # Enforce that all "public" methods of the Builder class have
        # 'filename' and 'lineno' as the first two arguments.
        for name, func in inspect.getmembers(grammar.Builder, inspect.isfunction):
            if (name.isupper() or
                name.startswith('_') or
                name.startswith('get_') or
                name == 'finalize'):
                continue
            parameters = inspect.signature(func).parameters.keys()
            self.assertEqual(list(parameters)[:3], ['self', 'filename', 'lineno'])


if __name__ == '__main__':
    unittest.main()



"""Fetch prices from Yahoo Finance's CSV API.

As of late 2017, the older Yahoo finance API deprecated. In particular, the
ichart endpoint is gone, and the download endpoint requires a cookie (which
could be gotten - here's some documentation for that
http://blog.bradlucas.com/posts/2017-06-02-new-yahoo-finance-quote-download-url/).

We're using both the v7 and v8 APIs here, both of which are, as far as I can
tell, undocumented:

https://query1.finance.yahoo.com/v7/finance/quote
https://query1.finance.yahoo.com/v8/finance/chart/SYMBOL

Timezone information: Input and output datetimes are specified via UNIX
timestamps, but the timezone of the particular market is included in the output.
"""
__copyright__ = "Copyright (C) 2015-2018  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
from typing import Dict, Any

import requests

from beancount.core.number import D
from beancount.prices import source


class YahooError(ValueError):
    "An error from the Yahoo API."


def parse_response(response: requests.models.Response) -> Dict:
    """Process as response from Yahoo.

    Raises:
      YahooError: If there is an error in the response.
    """
    json = response.json(parse_float=D)
    content = next(iter(json.values()))
    if response.status_code != requests.codes.ok:
        raise YahooError("Status {}: {}".format(response.status_code, content['error']))
    if len(json) != 1:
        raise YahooError("Invalid format in response from Yahoo; many keys: {}".format(
            ','.join(json.keys())))
    if content['error'] is not None:
        raise YahooError("Error fetching Yahoo data: {}".format(content['error']))
    return content['result'][0]


# Note: Feel free to suggest more here via a PR.
_MARKETS = {
    'us_market': 'USD',
    'ca_market': 'CAD',
}


def parse_currency(result: Dict[str, Any]) -> str:
    """Infer the currency from the result."""
    if 'market' not in result:
        return None
    return _MARKETS.get(result['market'], None)


_DEFAULT_PARAMS = {
    'lang': 'en-US',
    'corsDomain': 'finance.yahoo.com',
    '.tsrc': 'finance',
}


class Source(source.Source):
    "Yahoo Finance CSV API price extractor."

    def get_latest_price(self, ticker):
        """See contract in beancount.prices.source.Source."""

        url = "https://query1.finance.yahoo.com/v7/finance/quote"
        fields = ['symbol', 'regularMarketPrice', 'regularMarketTime']
        payload = {
            'symbols': ticker,
            'fields': ','.join(fields),
            'exchange': 'NYSE',
        }
        payload.update(_DEFAULT_PARAMS)
        response = requests.get(url, params=payload)
        result = parse_response(response)
        try:
            price = D(result['regularMarketPrice'])

            timezone = datetime.timezone(
                datetime.timedelta(hours=result['gmtOffSetMilliseconds'] / 3600000),
                result['exchangeTimezoneName'])
            trade_time = datetime.datetime.fromtimestamp(result['regularMarketTime'],
                                                         tz=timezone)
        except KeyError:
            raise YahooError("Invalid response from Yahoo: {}".format(repr(result)))

        currency = parse_currency(result)

        return source.SourcePrice(price, trade_time, currency)

    def get_historical_price(self, ticker, time):
        """See contract in beancount.prices.source.Source."""
        if requests is None:
            raise YahooError("You must install the 'requests' library.")
        url = "https://query1.finance.yahoo.com/v8/finance/chart/{}".format(ticker)
        dt_start = time - datetime.timedelta(days=5)
        dt_end = time
        payload = {
            'period1': int(dt_start.timestamp()),
            'period2': int(dt_end.timestamp()),
            'interval': '1d',
        }
        payload.update(_DEFAULT_PARAMS)
        response = requests.get(url, params=payload)
        result = parse_response(response)

        meta = result['meta']
        timezone = datetime.timezone(datetime.timedelta(hours=meta['gmtoffset'] / 3600),
                                     meta['exchangeTimezoneName'])

        timestamp_array = result['timestamp']
        close_array = result['indicators']['quote'][0]['close']
        series = [(datetime.datetime.fromtimestamp(timestamp, tz=timezone), D(price))
                  for timestamp, price in zip(timestamp_array, close_array)]

        # Get the latest data returned.
        latest = None
        for data_dt, price in sorted(series):
            if data_dt >= time:
                break
            latest = data_dt, price
        if latest is None:
            raise YahooError("Could not find price before {} in {}".format(time, series))

        currency = result['meta']['currency']
        return source.SourcePrice(price, data_dt, currency)



"""Full (new) booking implementation.

Problem description:

Interpolation and booking feed on each other, that is, numbers filled in from
interpolation might affect the booking process, and numbers derived from the
booking process may help carry out interpolation that would otherwise be
under-defined. Here's an example of interpolation helping the booking process:

Assume the ante-inventory of Assets:Investments contains two lots of shares of
HOOL, one at 100.00 USD and the other at 101.00 USD and apply this transaction:

    2015-09-30 *
      Assets:Investments   -10 HOOL {USD}
      Assets:Cash               1000 USD
      Income:Gains              -200 USD

Interpolation is unambiguously able to back out a cost of 100 USD / HOOL, which
would then result in an unambiguous booking result.

On the other hand, consider this transaction:

    2015-09-30 *
      Assets:Investments    -10 HOOL {USD}
      Assets:Cash               1000 USD
      Income:Gains

Now the interpolation cannot succeed. If the Assets:Investments account is
configured to use the FIFO method, the 10 oldest shares would be selected for
the cost, and we could then interpolate the capital gains correctly.

First observation: The second case is much more frequent than the first, and the
first is easily resolved manually by requiring a particular cost be specified.
Moreover, in many cases there isn't just a single lot of shares to be reduced
from and figuring out the correct set of shares given a target cost is an
underspecified problem.

Second observation: Booking can only be achieved for inventory reductions, not
for augmentations. Therefore, we should carry out booking on inventory
reductions and fail early if reduction is undefined there, and leave inventory
augmentations with missing numbers undefined, so that interpolation can fill
them in at a later stage.

Note that one case we'd like to but may not be able to handle is of a reduction
with interpolated price, like this:

    2015-09-30 *
      Assets:Investments        -10 HOOL {100.00 # USD}
      Expenses:Commission      9.95 USD
      Assets:Cash            990.05 USD

Therefore we choose to

1) Carry out booking first, on inventory reductions only, and leave inventory
   augmentations as they are, possibly undefined. The 'cost' attributed of
   booked postings are converted from CostSpec to Cost. Augmented postings with
   missing amounts are left as CostSpec instances in order to allow for
   interpolation of total vs. per-unit amount.

2) Compute interpolations on the resulting postings. Undefined costs for
   inventory augmentations may be filled in by interpolations at this stage (if
   possible).

3) Finally, convert the interpolated CostSpec instances to Cost instances.

Improving on this algorithm would require running a loop over the booking and
interpolation steps until all numbers are resolved or no more inference can
occur. We may consider that for later, as an experimental feature. My hunch is
that there are so few cases for which this would be useful that we won't bother
improving on the algorithm above.
"""
__copyright__ = "Copyright (C) 2015-2017  Martin Blais"
__license__ = "GNU GPLv2"

import collections
import copy
import enum
from decimal import Decimal
from typing import Text
import uuid

from beancount.core.number import MISSING
from beancount.core.number import ZERO
from beancount.core.data import Transaction
from beancount.core.data import Booking
from beancount.core.amount import Amount
from beancount.core.position import Position
from beancount.core.position import Cost
from beancount.core.position import CostSpec
from beancount.parser import booking_method
from beancount.core import position
from beancount.core import inventory
from beancount.core import interpolate


def unique_label() -> Text:
    "Return a globally unique label for cost entries."
    return str(uuid.uuid4())


# An error of disallowed self-reduction.
SelfReduxError = collections.namedtuple('SelfReduxError', 'source message entry')


def book(entries, options_map, methods):
    """Interpolate missing data from the entries using the full historical algorithm.
    See the internal implementation _book() for details.
    This method only stripes some of the return values.

    See _book() for arguments and return values.
    """
    entries, errors, _ = _book(entries, options_map, methods)
    return entries, errors


def _book(entries, options_map, methods):
    """Interpolate missing data from the entries using the full historical algorithm.

    Args:
      incomplete_entries: A list of directives, with some postings possibly left
        with incomplete amounts as produced by the parser.
      options_map: An options dict as produced by the parser.
      methods: A mapping of account name to their corresponding booking
        method.
    Returns:
      A triple of
        entries: A list of interpolated entries with all their postings completed.
        errors: New errors produced during interpolation.
        balances: A dict of account name and resulting balances.
    """
    new_entries = []
    errors = []
    balances = collections.defaultdict(inventory.Inventory)
    for entry in entries:
        if isinstance(entry, Transaction):
            # Group postings by currency.
            refer_groups, cat_errors = categorize_by_currency(entry, balances)
            if cat_errors:
                errors.extend(cat_errors)
                continue
            posting_groups = replace_currencies(entry.postings, refer_groups)

            # Get the list of tolerances.
            tolerances = interpolate.infer_tolerances(entry.postings, options_map)

            # Resolve reductions to a particular lot in their inventory balance.
            repl_postings = []
            for currency, group_postings in posting_groups:
                # Important note: the group of 'postings' here is a subset of
                # that from entry.postings, and may include replicated
                # auto-postings. Never use entry.postings going forward.

                # (See http://furius.ca/beancount/doc/self-reductions for an
                # explanation of how we will eventually treat each currency
                # group in this block; Summary: We will need to run the
                # reductions prior to the augmentations in order to support
                # reductions between the postings of a single transaction.)
                # Disabled.
                if False:  # pylint: disable=using-constant-test
                    if has_self_reduction(group_postings, methods):
                        errors.append(SelfReduxError(
                            entry.meta, "Self-reduction is not allowed", entry))

                # Perform booking reductions, that is, match postings which
                # reduce the ante-inventory of their accounts to an existing
                # position in the inventory against a possibly incomplete
                # CostSpec specification, and replace the postings' cost to the
                # fully-specified (with a date & label) existing Cost instance.
                # Note that 'balances' remains untouched.
                #
                # Also note that 'booked_postings' may include augmenting
                # postings whose 'cost' attribute has been left to a CostSpec
                # instance. Therefore, the postings held-at-cost may hold a
                # mixture of Cost and CostSpec instances. This is necessary to
                # let the interpolation do its magic on partially incomplete
                # CostSpec instances below.
                (booked_postings,
                 booking_errors) = book_reductions(entry, group_postings, balances,
                                                   methods)

                # If there were any errors, skip this group of postings.
                if booking_errors:
                    errors.extend(booking_errors)
                    continue

                # Interpolate missing numbers from all postings. This
                # includes partially incomplete CostSpec instances remaining
                # on augmenting postings. After this interpolation, all
                # 'inter_postings' consists entirely of postings holding
                # instances of Cost.
                (inter_postings,
                 interpolation_errors,
                 interpolated) = interpolate_group(booked_postings, balances, currency,
                                                   tolerances)

                if interpolation_errors:
                    errors.extend(interpolation_errors)
                repl_postings.extend(inter_postings)

            # Replace postings by interpolated ones.
            meta = entry.meta.copy()
            meta[interpolate.AUTOMATIC_TOLERANCES] = tolerances
            entry = entry._replace(postings=repl_postings,
                                   meta=meta)

            # Update the running balances for each account using the final,
            # booked and interpolated values. Note that we could optimize away
            # some of this in book_reductions() but we choose not to do so, as a
            # sanity check that the direct aggregation of the final booked lots
            # will compute the same result as that during the book_reductions()
            # process.
            for posting in repl_postings:
                balance = balances[posting.account]
                balance.add_position(posting)

        new_entries.append(entry)

    return new_entries, errors, balances


# An error raised if we failed to bucket a posting to a particular currency.
CategorizationError = collections.namedtuple('CategorizationError', 'source message entry')


def get_bucket_currency(refer):
    """Given currency references for a posting, return the bucket currency.

    Args:
      refer: An instance of Refer.
    Returns:
      A currency string.
    """
    currency = None
    if isinstance(refer.cost_currency, str):
        currency = refer.cost_currency
    elif isinstance(refer.price_currency, str):
        currency = refer.price_currency
    elif (refer.cost_currency is None and
          refer.price_currency is None and
          isinstance(refer.units_currency, str)):
        currency = refer.units_currency
    return currency

Refer = collections.namedtuple('Refer', 'index units_currency cost_currency price_currency')


def categorize_by_currency(entry, balances):
    """Group the postings by the currency they declare.

    This is used to prepare the postings for the next stages: Interpolation and
    booking will then be carried out separately on each currency group. At the
    outset of this routine, we should have distinct groups of currencies without
    any ambiguities regarding which currency they need to balance against.

    Here's how this works.

    - First we apply the constraint that cost-currency and price-currency must
      match, if there is both a cost and a price. This reduces the space of
      possibilities somewhat.

    - If the currency is explicitly specified, we put the posting in that
      currency's bucket.

    - If not, we have a few methods left to disambiguate the currency:

      1. We look at the remaining postings... if they are all of a single
         currency, the posting must be in that currency too.

      2. If we cannot do that, we inspect the contents of the inventory of the
         account for the posting. If all the contents are of a single currency,
         we use that one.

    Args:
      postings: A list of incomplete postings to categorize.
      balances: A dict of currency to inventory contents before the transaction is
        applied.
    Returns:
      A list of (currency string, list of tuples) items describing each postings
      and its interpolated currencies, and a list of generated errors for
      currency interpolation. The entry's original postings are left unmodified.
      Each tuple in the value-list contains:
        index: The posting index in the original entry.
        units_currency: The interpolated currency for units.
        cost_currency: The interpolated currency for cost.
        price_currency: The interpolated currency for price.
    """
    errors = []

    groups = collections.defaultdict(list)
    sortdict = {}
    auto_postings = []
    unknown = []
    for index, posting in enumerate(entry.postings):
        units = posting.units
        cost = posting.cost
        price = posting.price

        # Extract and override the currencies locally.
        units_currency = (units.currency
                          if units is not MISSING and units is not None
                          else None)
        cost_currency = (cost.currency
                         if cost is not MISSING and cost is not None
                         else None)
        price_currency = (price.currency
                          if price is not MISSING and price is not None
                          else None)

        # First we apply the constraint that cost-currency and price-currency
        # must match, if there is both a cost and a price. This reduces the
        # space of possibilities somewhat.
        if cost_currency is MISSING and isinstance(price_currency, str):
            cost_currency = price_currency
        if price_currency is MISSING and isinstance(cost_currency, str):
            price_currency = cost_currency

        refer = Refer(index, units_currency, cost_currency, price_currency)

        if units is MISSING and price_currency is None:
            # Bucket auto-postings separately from unknown.
            auto_postings.append(refer)
        else:
            # Bucket with what we know so far.
            currency = get_bucket_currency(refer)
            if currency is not None:
                sortdict.setdefault(currency, index)
                groups[currency].append(refer)
            else:
                # If we need to infer the currency, store in unknown.
                unknown.append(refer)

    # We look at the remaining postings... if they are all of a single currency,
    # the posting must be in that currency too.
    if unknown and len(unknown) == 1 and len(groups) == 1:
        (index, units_currency, cost_currency, price_currency) = unknown.pop()

        other_currency = next(iter(groups.keys()))
        if price_currency is None and cost_currency is None:
            # Infer to the units currency.
            units_currency = other_currency
        else:
            # Infer to the cost and price currencies.
            if price_currency is MISSING:
                price_currency = other_currency
            if cost_currency is MISSING:
                cost_currency = other_currency

        refer = Refer(index, units_currency, cost_currency, price_currency)
        currency = get_bucket_currency(refer)
        assert currency is not None
        sortdict.setdefault(currency, index)
        groups[currency].append(refer)

    # Finally, try to resolve all the unknown legs using the inventory contents
    # of each account.
    for refer in unknown:
        (index, units_currency, cost_currency, price_currency) = refer
        posting = entry.postings[index]
        balance = balances.get(posting.account, None)
        if balance is None:
            balance = inventory.Inventory()

        if units_currency is MISSING:
            balance_currencies = balance.currencies()
            if len(balance_currencies) == 1:
                units_currency = balance_currencies.pop()

        if cost_currency is MISSING or price_currency is MISSING:
            balance_cost_currencies = balance.cost_currencies()
            if len(balance_cost_currencies) == 1:
                balance_cost_currency = balance_cost_currencies.pop()
                if price_currency is MISSING:
                    price_currency = balance_cost_currency
                if cost_currency is MISSING:
                    cost_currency = balance_cost_currency

        refer = Refer(index, units_currency, cost_currency, price_currency)
        currency = get_bucket_currency(refer)
        if currency is not None:
            sortdict.setdefault(currency, index)
            groups[currency].append(refer)
        else:
            errors.append(
                CategorizationError(posting.meta,
                                    "Failed to categorize posting {}".format(index + 1),
                                    entry))

    # Fill in missing units currencies if some remain as missing. This may occur
    # if we used the cost or price to bucket the currency but the units currency
    # was missing.
    for currency, refers in groups.items():
        for rindex, refer in enumerate(refers):
            if refer.units_currency is MISSING:
                posting = entry.postings[refer.index]
                balance = balances.get(posting.account, None)
                if balance is None:
                    continue
                balance_currencies = balance.currencies()
                if len(balance_currencies) == 1:
                    refers[rindex] = refer._replace(units_currency=balance_currencies.pop())

    # Deal with auto-postings.
    if len(auto_postings) > 1:
        refer = auto_postings[-1]
        posting = entry.postings[refer.index]
        errors.append(
            CategorizationError(posting.meta,
                                "You may not have more than one auto-posting per currency",
                                entry))
        auto_postings = auto_postings[0:1]
    for refer in auto_postings:
        for currency in groups.keys():
            sortdict.setdefault(currency, refer.index)
            groups[currency].append(Refer(refer.index, currency, None, None))

    # Issue error for all currencies which we could not resolve.
    for currency, refers in groups.items():
        for refer in refers:
            posting = entry.postings[refer.index]
            for currency, name in [(refer.units_currency, 'units'),
                                   (refer.cost_currency, 'cost'),
                                   (refer.price_currency, 'price')]:
                if currency is MISSING:
                    errors.append(CategorizationError(
                        posting.meta,
                        "Could not resolve {} currency".format(name),
                        entry))

    sorted_groups = sorted(groups.items(), key=lambda item: sortdict[item[0]])
    return sorted_groups, errors


def replace_currencies(postings, refer_groups):
    """Replace resolved currencies in the entry's Postings.

    This essentially applies the findings of categorize_by_currency() to produce
    new postings with all currencies resolved.

    Args:
      postings: A list of Posting instances to replace.
      refer_groups: A list of (currency, list of posting references) items as
        returned by categorize_by_currency().
    Returns:
      A new list of items of (currency, list of Postings), postings for which the
      currencies have been replaced by their interpolated currency values.
    """
    new_groups = []
    for currency, refers in refer_groups:
        new_postings = []
        for refer in sorted(refers, key=lambda r: r.index):
            posting = postings[refer.index]
            units = posting.units
            if units is MISSING or units is None:
                posting = posting._replace(units=Amount(MISSING, refer.units_currency))
            else:
                replace = False
                cost = posting.cost
                price = posting.price
                if units.currency is MISSING:
                    units = Amount(units.number, refer.units_currency)
                    replace = True
                if cost and cost.currency is MISSING:
                    cost = cost._replace(currency=refer.cost_currency)
                    replace = True
                if price and price.currency is MISSING:
                    price = Amount(price.number, refer.price_currency)
                    replace = True
                if replace:
                    posting = posting._replace(units=units, cost=cost, price=price)
            new_postings.append(posting)
        new_groups.append((currency, new_postings))
    return new_groups


# An error raised if we failed to reduce the inventory balance unambiguously.
ReductionError = collections.namedtuple('ReductionError', 'source message entry')


def has_self_reduction(postings, methods):
    """Return true if the postings potentially reduce each other at cost.

    Args:
      postings: A list of postings with uninterpolated CostSpec cost instances.
      methods: A mapping of account name to their corresponding booking
        method.
    Returns:
      A boolean, true if there's a potential for self-reduction.
    """
    # A mapping of (currency, cost-currency) and sign.
    cost_changes = {}
    for posting in postings:
        cost = posting.cost
        if cost is None:
            continue
        if methods[posting.account] is Booking.NONE:
            continue
        key = (posting.account, posting.units.currency)
        sign = 1 if posting.units.number > ZERO else -1
        if cost_changes.setdefault(key, sign) != sign:
            return True
    return False


def book_reductions(entry, group_postings, balances,
                    methods):
    """Book inventory reductions against the ante-balances.

    This function accepts a dict of (account, Inventory balance) and for each
    posting that is a reduction against its inventory, attempts to find a
    corresponding lot or list of lots to reduce the balance with.

    * For reducing lots, the CostSpec instance of the posting is replaced by a
      Cost instance.

    * For augmenting lots, the CostSpec instance of the posting is left alone,
      except for its date, which is inherited from the parent Transaction.

    Args:
      entry: An instance of Transaction. This is only used to refer to when
        logging errors.
      group_postings: A list of Posting instances for the group.
      balances: A dict of account name to inventory contents.
      methods: A mapping of account name to their corresponding booking
        method enum.
    Returns:
      A pair of
        booked_postings: A list of booked postings, with reducing lots resolved
          against specific position in the corresponding accounts'
          ante-inventory balances. Note single reducing posting in the input may
          result in multiple postings in the output. Also note that augmenting
          postings held-at-cost will still refer to 'cost' instances of
          CostSpec, left to be interpolated later.
        errors: A list of errors, if there were any.
    """
    errors = []

    # A local copy of the balances dictionary which is updated just for the
    # duration of this function's updates, in order to take into account the
    # cumulative effect of all the postings inferred here
    local_balances = {}

    empty = inventory.Inventory()
    booked_postings = []
    for posting in group_postings:
        # Process a single posting.
        units = posting.units
        costspec = posting.cost
        account = posting.account

        # Note: We ensure there is no mutation on 'balances' to keep this
        # function without side-effects. Note that we may be able to optimize
        # performance later on by giving up this property.
        #
        # Also note that if there is no existing balance, then won't be any lot
        # reduction because none of the postings will be able to match against
        # any currencies of the balance.
        previous_balance = balances.get(account, empty)
        balance = local_balances.setdefault(account, copy.copy(previous_balance))

        # Check if this is a lot held at cost.
        if costspec is None or units.number is MISSING:
            # This posting is not held at cost; we do nothing.
            booked_postings.append(posting)
        else:
            # This posting is held at cost; figure out if it's a reduction or an
            # augmentation.
            method = methods[account]
            if (method is not Booking.NONE and
                balance is not None and
                balance.is_reduced_by(units)):
                # This posting is a reduction.

                # Match the positions.
                cost_number = compute_cost_number(costspec, units)
                matches = []
                for position in balance:
                    # Skip inventory contents of a different currency.
                    if (units.currency and
                        position.units.currency != units.currency):
                        continue
                    # Skip balance positions not held at cost.
                    if position.cost is None:
                        continue
                    if (cost_number is not None and
                        position.cost.number != cost_number):
                        continue
                    if (isinstance(costspec.currency, str) and
                        position.cost.currency != costspec.currency):
                        continue
                    if (costspec.date and
                        position.cost.date != costspec.date):
                        continue
                    if (costspec.label and
                        position.cost.label != costspec.label):
                        continue
                    matches.append(position)

                # Check for ambiguous matches.
                if len(matches) == 0:
                    errors.append(
                        ReductionError(entry.meta,
                                       'No position matches "{}" against balance {}'.format(
                                           posting, balance),
                                       entry))
                    return [], errors  # This is irreconcilable, remove these postings.

                reduction_postings, matched_postings, ambi_errors = (
                    booking_method.handle_ambiguous_matches(entry, posting, matches,
                                                            method))
                if ambi_errors:
                    errors.extend(ambi_errors)
                    return [], errors

                # Add the reductions to the resulting list of booked postings.
                booked_postings.extend(reduction_postings)

                # Update the local balance in order to avoid matching against
                # the same postings twice when processing multiple postings in
                # the same transaction. Note that we only do this for postings
                # held at cost because the other postings may need interpolation
                # in order to be resolved properly.
                for posting in reduction_postings:
                    balance.add_position(posting)
            else:
                # This posting is an augmentation.
                #
                # Note that we do not convert the CostSpec instances to Cost
                # instances, because we want to let the subsequent interpolation
                # process able to interpolate either the cost per-unit or the
                # total cost, separately.

                # Put in the date of the parent Transaction if there is no
                # explicit date specified on the spec.
                if costspec.date is None:
                    dated_costspec = costspec._replace(date=entry.date)
                    posting = posting._replace(cost=dated_costspec)

                # FIXME: Insert unique ids for trade tracking; right now this
                # creates ambiguous matches errors (and it shouldn't).
                # # Insert a unique label if there isn't one.
                # if posting.cost is not None and posting.cost.label is None:
                #     posting = posting._replace(
                #         cost=posting.cost._replace(label=unique_label()))

                booked_postings.append(posting)

    return booked_postings, errors


def compute_cost_number(costspec, units):
    """Given a CostSpec, return the cost number, if possible to compute.

    Args:
      costspec: A parsed instance of CostSpec.
      units: An instance of Amount for the units of the position.
    Returns:
      If it is not possible to calculate the cost, return None.
      Otherwise, returns a Decimal instance, the per-unit cost.
    """
    number_per = costspec.number_per
    number_total = costspec.number_total
    if MISSING in (number_per, number_total):
        return None
    if number_total is not None:
        # Compute the per-unit cost if there is some total cost
        # component involved.
        cost_total = number_total
        units_number = units.number
        if number_per is not None:
            cost_total += number_per * units_number
        unit_cost = cost_total / abs(units_number)
    elif number_per is None:
        return None
    else:
        unit_cost = number_per
    return unit_cost


def convert_costspec_to_cost(posting):
    """Convert an instance of CostSpec to Cost, if present on the posting.

    If the posting has no cost, it itself is just returned.

    Args:
      posting: An instance of Posting.
    Returns:
      An instance of Posting with a possibly replaced 'cost' attribute.
    """
    cost = posting.cost
    if isinstance(cost, position.CostSpec):
        if cost is not None:
            units_number = posting.units.number
            number_per = cost.number_per
            number_total = cost.number_total
            if number_total is not None:
                # Compute the per-unit cost if there is some total cost
                # component involved.
                cost_total = number_total
                if number_per is not MISSING:
                    cost_total += number_per * units_number
                unit_cost = cost_total / abs(units_number)
            else:
                unit_cost = number_per
            new_cost = Cost(unit_cost, cost.currency, cost.date, cost.label)
            posting = posting._replace(units=posting.units, cost=new_cost)
    return posting


# FIXME: Refactor compute_cost_number() and convert_costspec_to_cost().


class MissingType(enum.Enum):
    """The type of missing number."""
    UNITS = 1
    COST_PER = 2
    COST_TOTAL = 3
    PRICE = 4


# An error raised if we are not able to interpolate.
InterpolationError = collections.namedtuple('InterpolationError', 'source message entry')


def interpolate_group(postings, balances, currency, tolerances):
    """Interpolate missing numbers in the set of postings.

    Args:
      postings: A list of Posting instances.
      balances: A dict of account to its ante-inventory.
      currency: The weight currency of this group, used for reporting errors.
      tolerances: A dict of currency to tolerance values.
    Returns:
      A tuple of
        postings: A lit of new posting instances.
        errors: A list of errors generated during interpolation.
        interpolated: A boolean, true if we did have to interpolate.

      In the case of an error, this returns the original list of postings, which
      is still incomplete. If an error is returned, you should probably skip the
      transaction altogether, or just not include the postings in it. (An
      alternative behaviour would be to return only the list of valid postings,
      but that would likely result in an unbalanced transaction. We do it this
      way by choice.)
    """
    errors = []

    # Figure out which type of amount is missing, by creating a list of
    # incomplete postings and which type of units is missing.
    incomplete = []
    for index, posting in enumerate(postings):
        units = posting.units
        cost = posting.cost
        price = posting.price

        # Identify incomplete parts of the Posting components.
        if units.number is MISSING:
            incomplete.append((MissingType.UNITS, index))

        if isinstance(cost, CostSpec):
            if cost and cost.number_per is MISSING:
                incomplete.append((MissingType.COST_PER, index))
            if cost and cost.number_total is MISSING:
                incomplete.append((MissingType.COST_TOTAL, index))
        else:
            # Check that a resolved instance of Cost never needs interpolation.
            #
            # Note that in theory we could support the interpolation of regular
            # per-unit costs in these if we wanted to; but because they're all
            # reducing postings that have been booked earlier, those never need
            # to be interpolated.
            if cost is not None:
                assert isinstance(cost.number, Decimal), (
                    "Internal error: cost has no number: {}".format(cost))

        if price and price.number is MISSING:
            incomplete.append((MissingType.PRICE, index))

    # The replacement posting for the incomplete posting of this group.
    new_posting = None

    if len(incomplete) == 0:
        # If there are no missing numbers, just convert the CostSpec to Cost and
        # return that.
        out_postings = [convert_costspec_to_cost(posting)
                        for posting in postings]

    elif len(incomplete) > 1:
        # If there is more than a single value to be interpolated, generate an
        # error and return no postings.
        _, posting_index = incomplete[0]
        errors.append(InterpolationError(
            postings[posting_index].meta,
            "Too many missing numbers for currency group '{}'".format(currency),
            None))
        out_postings = []

    else:
        # If there is a single missing number, calculate it and fill it in here.
        missing, index = incomplete[0]
        incomplete_posting = postings[index]

        # Convert augmenting postings' costs from CostSpec to corresponding Cost
        # instances, except for the incomplete posting.
        new_postings = [(posting
                         if posting is incomplete_posting
                         else convert_costspec_to_cost(posting))
                        for posting in postings]

        # Compute the balance of the other postings.
        residual = interpolate.compute_residual(posting
                                                for posting in new_postings
                                                if posting is not incomplete_posting)
        assert len(residual) < 2, "Internal error in grouping postings by currencies."
        if not residual.is_empty():
            respos = next(iter(residual))
            assert respos.cost is None, (
                "Internal error; cost appears in weight calculation.")
            assert respos.units.currency == currency, (
                "Internal error; residual different than currency group.")
            weight = -respos.units.number
            weight_currency = respos.units.currency
        else:
            weight = ZERO
            weight_currency = currency

        if missing == MissingType.UNITS:
            units = incomplete_posting.units
            cost = incomplete_posting.cost
            if cost:
                # Handle the special case where we only have total cost.
                if cost.number_per == ZERO:
                    errors.append(InterpolationError(
                        incomplete_posting.meta,
                        "Cannot infer per-unit cost only from total", None))
                    return postings, errors, True

                assert cost.currency == weight_currency, (
                    "Internal error; residual currency different than missing currency.")
                cost_total = cost.number_total or ZERO
                units_number = (weight - cost_total) / cost.number_per

            elif incomplete_posting.price:
                assert incomplete_posting.price.currency == weight_currency, (
                    "Internal error; residual currency different than missing currency.")
                units_number = weight / incomplete_posting.price.number

            else:
                assert units.currency == weight_currency, (
                    "Internal error; residual currency different than missing currency.")
                units_number = weight

            # Quantize the interpolated units if necessary.
            units_number = interpolate.quantize_with_tolerance(tolerances,
                                                               units.currency,
                                                               units_number)

            if weight != ZERO:
                new_pos = Position(Amount(units_number, units.currency), cost)
                new_posting = incomplete_posting._replace(units=new_pos.units,
                                                          cost=new_pos.cost)
            else:
                new_posting = None

        elif missing == MissingType.COST_PER:
            units = incomplete_posting.units
            cost = incomplete_posting.cost
            assert cost.currency == weight_currency, (
                "Internal error; residual currency different than missing currency.")
            if units.number != ZERO:
                number_per = (weight - (cost.number_total or ZERO)) / units.number
                new_cost = cost._replace(number_per=number_per)
                new_pos = Position(units, new_cost)
                new_posting = incomplete_posting._replace(units=new_pos.units,
                                                          cost=new_pos.cost)
            else:
                new_posting = None

        elif missing == MissingType.COST_TOTAL:
            units = incomplete_posting.units
            cost = incomplete_posting.cost
            assert cost.currency == weight_currency, (
                "Internal error; residual currency different than missing currency.")
            number_total = (weight - cost.number_per * units.number)
            new_cost = cost._replace(number_total=number_total)
            new_pos = Position(units, new_cost)
            new_posting = incomplete_posting._replace(units=new_pos.units,
                                                      cost=new_pos.cost)

        elif missing == MissingType.PRICE:
            units = incomplete_posting.units
            cost = incomplete_posting.cost
            if cost is not None:
                errors.append(InterpolationError(
                    incomplete_posting.meta,
                    "Cannot infer price for postings with units held at cost", None))
                return postings, errors, True
            else:
                price = incomplete_posting.price
                assert price.currency == weight_currency, (
                    "Internal error; residual currency different than missing currency.")
                new_price_number = abs(weight / units.number)
                new_posting = incomplete_posting._replace(price=Amount(new_price_number,
                                                                       price.currency))

        else:
            assert False, "Internal error; Invalid missing type."

        # Replace the number in the posting.
        if new_posting is not None:
            # Set meta-data on the new posting to indicate it was interpolated.
            if new_posting.meta is None:
                new_posting = new_posting._replace(meta={})
            new_posting.meta[interpolate.AUTOMATIC_META] = True

            # Convert augmenting posting costs from CostSpec to a corresponding
            # Cost instance.
            new_postings[index] = convert_costspec_to_cost(new_posting)
        else:
            del new_postings[index]
        out_postings = new_postings

    assert all(not isinstance(posting.cost, CostSpec)
               for posting in out_postings)

    # Check that units are non-zero and that no cost remains negative; issue an
    # error if this is the case.
    for posting in out_postings:
        if posting.cost is None:
            continue
        # If there is a cost, we don't allow either a cost value of zero,
        # nor a zero number of units. Note that we allow a price of zero as
        # the only special case allowed (for conversion entries), but never
        # for costs.
        if posting.units.number == ZERO:
            errors.append(InterpolationError(
                posting.meta,
                'Amount is zero: "{}"'.format(posting.units), None))
        if posting.cost.number < ZERO:
            errors.append(InterpolationError(
                posting.meta,
                'Cost is negative: "{}"'.format(posting.cost), None))

    return out_postings, errors, (new_posting is not None)



__copyright__ = "Copyright (C) 2014, 2016-2017  Martin Blais"
__license__ = "GNU GPLv2"

import unittest

from beancount.web import bottle_utils


class TestAttrMapper(unittest.TestCase):

    def setUp(self):
        dictionary = {'a': 'the_a',
                      'b': 'the_b',
                      'c': 'the_c'}
        self.mapper = bottle_utils.AttrMapper(dictionary.__getitem__)

    def test_AttrMapper(self):
        self.assertTrue(self.mapper is not None)

    def test_AttrMapper_getattr(self):
        self.assertTrue(self.mapper.a, 'the_a')
        self.assertTrue(self.mapper.b, 'the_b')

    def test_AttrMapper_build(self):
        self.assertTrue(self.mapper.build('a'), 'the_a')
        self.assertTrue(self.mapper.build('b'), 'the_b')


if __name__ == '__main__':
    unittest.main()



__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import tempfile
import shutil
import unittest
from unittest import mock

from beancount.ingest import cache
from beancount.utils import file_type


class TestFileMemo(unittest.TestCase):

    def test_cache(self):
        with tempfile.NamedTemporaryFile() as tmpfile:
            shutil.copy(__file__, tmpfile.name)
            wrap = cache._FileMemo(tmpfile.name)

            # Check attributes.
            self.assertEqual(tmpfile.name, wrap.name)

            # Check that caching works.
            converter = mock.MagicMock(return_value='abc')
            self.assertEqual('abc', wrap.convert(converter))
            self.assertEqual('abc', wrap.convert(converter))
            self.assertEqual('abc', wrap.convert(converter))
            self.assertEqual(1, converter.call_count)

    @unittest.skipIf(not file_type.magic, 'python-magic is not installed')
    def test_cache_head_and_contents(self):
        with tempfile.NamedTemporaryFile() as tmpfile:
            shutil.copy(__file__, tmpfile.name)
            wrap = cache._FileMemo(tmpfile.name)

            contents = wrap.convert(cache.contents)
            self.assertIsInstance(contents, str)
            self.assertGreater(len(contents), 128)

            contents2 = wrap.contents()
            self.assertEqual(contents, contents2)

            head = wrap.convert(cache.head(128))
            self.assertIsInstance(head, str)
            self.assertEqual(128, len(head))

            mimetype = wrap.convert(cache.mimetype)
            self.assertRegex(mimetype, r'text/(x-(python|c\+\+)|plain)')


if __name__ == '__main__':
    unittest.main()



"""A plugin of plugins which triggers are all the automatic and lax plugins.

In a sense, this is the inverse of "pedantic." This is useful when doing some
types of quick and dirty tests. You can just import the "auto" plugin. Put that
in a macro.

Also see: the 'pedantic' plugin.
"""
__copyright__ = "Copyright (C) 2017  Martin Blais"
__license__ = "GNU GPLv2"

from beancount.plugins import auto_accounts
from beancount.plugins import implicit_prices
from beancount import loader

__plugins__ = loader.combine_plugins(
    auto_accounts,
    implicit_prices)



"""Bake a Beancount input file's web files to a directory hierarchy.

You provide a Beancount filename, an output directory, and this script
runs a server and a scraper that puts all the files in the directory,
and if your output name has an archive suffix, we automatically the
fetched directory contents to the archive and delete them.
"""
__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

from os import path
import functools
import importlib
import logging
import os
import re
import shlex
import shutil
import subprocess
import zipfile

import lxml.html

from beancount.utils import scrape
from beancount.web import web
from beancount.utils import file_utils
from beancount.parser import version


# Directories where binary files are allowed.
BINARY_DIRECTORIES = ['resources', 'third_party', 'doc']
BINARY_MATCH = re.compile(r'/({}/|favicon.ico$)'.format(
    '|'.join(BINARY_DIRECTORIES))).match


def normalize_filename(url):
    """Convert URL paths to filenames. Add .html extension if needed.

    Args:
      url: A string, the url to convert.
    Returns:
      A string, possibly with an extension appended.
    """
    if url.endswith('/'):
        return path.join(url, 'index.html')
    elif BINARY_MATCH(url):
        return url
    else:
        return url if url.endswith('.html') else (url + '.html')


def relativize_links(html, current_url):
    """Make all the links in the contents string relative to an URL.

    Args:
      html: An lxml document node.
      current_url: A string, the URL of the current page, a path to.
        a file or a directory. If the path represents a directory, the
        path ends with a /.
    """
    current_dir = path.dirname(current_url)
    for element, attribute, link, pos in lxml.html.iterlinks(html):
        if path.isabs(link):
            relative_link = path.relpath(normalize_filename(link), current_dir)
            element.set(attribute, relative_link)


def remove_links(html, targets):
    """Convert a list of anchors (<a>) from an HTML tree to spans (<span>).

    Args:
      html: An lxml document node.
      targets: A set of string, targets to be removed.
    """
    for element, attribute, link, pos in lxml.html.iterlinks(html):
        if link in targets:
            del element.attrib[attribute]
            element.tag = 'span'
            element.set('class', 'removed-link')


def save_scraped_document(output_dir, url, response, contents, html_root, skipped_urls):
    """Callback function to process a document being scraped.

    This converts the document to have relative links and writes out the file to
    the output directory.

    Args:
      output_dir: A string, the output directory to write.
      url: A string, the originally requested URL.
      response: An http response as per urlopen.
      contents: Bytes, the content of a response.
      html_root: An lxml root node for the document, optionally. If this is provided,
        this avoid you having to reprocess it (for performance reasons).
      skipped_urls: A set of the links from the file that were skipped.
    """
    if response.status != 200:
        logging.error("Invalid status: %s", response.status)

    # Ignore directories.
    if url.endswith('/'):
        return

    # Note that we're saving the file under the non-redirected URL, because this
    # will have to be opened using files and there are no redirects that way.

    if response.info().get_content_type() == 'text/html':
        if html_root is None:
            html_root = lxml.html.document_fromstring(contents)
        remove_links(html_root, skipped_urls)
        relativize_links(html_root, url)
        contents = lxml.html.tostring(html_root, method="html")

    # Compute output filename and write out the relativized contents.
    output_filename = path.join(output_dir,
                                normalize_filename(url).lstrip('/'))
    os.makedirs(path.dirname(output_filename), exist_ok=True)
    with open(output_filename, 'wb') as outfile:
        outfile.write(contents)


def bake_to_directory(webargs, output_dir, render_all_pages=True):
    """Serve and bake a Beancount's web to a directory.

    Args:
      webargs: An argparse parsed options object with the web app arguments.
      output_dir: A directory name. We don't check here whether it exists or not.
      quiet: A boolean, True to suppress web server fetch log.
      render_all_pages: If true, fetch the full set of pages, not just the subset that
        is palatable.
    Returns:
      True on success, False otherwise.
    """
    callback = functools.partial(save_scraped_document, output_dir)

    if render_all_pages:
        ignore_regexps = None
    else:
        regexps = [
            # Skip the context pages, too slow.
            r'/context/',
            # Skip the link pages, too slow.
            r'/link/',
            # Skip the component pages... too many.
            r'/view/component/',
            # Skip served documents.
            r'/.*/doc/',
            # Skip monthly pages.
            r'/view/year/\d\d\d\d/month/',
        ]
        ignore_regexps = '({})'.format('|'.join(regexps))

    processed_urls, skipped_urls = web.scrape_webapp(webargs, callback, ignore_regexps)


def archive(command_template, directory, archive, quiet=False):
    """Archive the directory to the given tar/gz archive filename.

    Args:
      command_template: A string, the command template to format with in order
        to compute the command to run.
      directory: A string, the name of the directory to archive.
      archive: A string, the name of the file to output.
      quiet: A boolean, True to suppress output.
    Raises:
      IOError: if the directory does not exist or if the archive name already
      exists.

    """
    directory = path.abspath(directory)
    archive = path.abspath(archive)
    if not path.exists(directory):
        raise IOError("Directory to archive '{}' does not exist".format(
            directory))
    if path.exists(archive):
        raise IOError("Output archive name '{}' already exists".format(
            archive))

    command = command_template.format(directory=directory,
                                      dirname=path.dirname(directory),
                                      basename=path.basename(directory),
                                      archive=archive)

    pipe = subprocess.Popen(shlex.split(command),
                            shell=False,
                            cwd=path.dirname(directory),
                            stdout=subprocess.PIPE if quiet else None,
                            stderr=subprocess.PIPE if quiet else None)
    _, _ = pipe.communicate()
    if pipe.returncode != 0:
        raise OSError("Archive failure")


def archive_zip(directory, archive):
    """Archive the directory to the given tar/gz archive filename.

    Args:
      directory: A string, the name of the directory to archive.
      archive: A string, the name of the file to output.
    """
    # Figure out optimal level of compression among the supported ones in this
    # installation.
    for spec, compression in [
            ('lzma', zipfile.ZIP_LZMA),
            ('bz2', zipfile.ZIP_BZIP2),
            ('zlib', zipfile.ZIP_DEFLATED)]:
        if importlib.util.find_spec(spec):
            zip_compression = compression
            break
    else:
        # Default is no compression.
        zip_compression = zipfile.ZIP_STORED

    with file_utils.chdir(directory), zipfile.ZipFile(
            archive, 'w', compression=zip_compression) as archfile:
        for root, dirs, files in os.walk(directory):
            for filename in files:
                relpath = path.relpath(path.join(root, filename), directory)
                archfile.write(relpath)


ARCHIVERS = {
    '.tar.gz'  : 'tar -C {dirname} -zcvf {archive} {basename}',
    '.tgz'     : 'tar -C {dirname} -zcvf {archive} {basename}',
    '.tar.bz2' : 'tar -C {dirname} -jcvf {archive} {basename}',
    '.zip'     : archive_zip,
    }


def main():
    parser = version.ArgumentParser(description=__doc__)

    web_group = web.add_web_arguments(parser)
    web_group.set_defaults(port=9475)

    group = parser.add_argument_group("Bake process arguments")

    group.add_argument('output',
                       help=('The output directory or archive name. If you '
                             'specify a filename with a well-known extension,'
                             'we automatically archive the fetched directory '
                             'contents to this archive name and delete them.'))

    # In order to be able to bake in a reasonable amount of time, we need to
    # remove some pages; you can use this switch to do that.
    group.add_argument('--render-all-pages', '--full', action='store_true',
                       help=("Don't ignore some of the more numerious pages, "
                             "like monthly reports."))

    opts = parser.parse_args()

    # Figure out the archival method.
    output_directory, extension = file_utils.path_greedy_split(opts.output)
    if extension:
        try:
            archival_command = ARCHIVERS[extension]
        except KeyError:
            raise SystemExit("ERROR: Unknown archiver type '{}'".format(extension))
    else:
        archival_command = None

    # Check pre-conditions on input/output filenames.
    if not path.exists(opts.filename):
        raise SystemExit("ERROR: Missing input file '{}'".format(opts.filename))
    if path.exists(opts.output):
        raise SystemExit("ERROR: Output path already exists '{}'".format(opts.output))
    if path.exists(output_directory):
        raise SystemExit(
            "ERROR: Output directory already exists '{}'".format(output_directory))

    # Bake to a directory hierarchy of files with local links.
    bake_to_directory(opts, output_directory, opts.render_all_pages)

    # Verify the bake output files. This is just a sanity checking step.
    # You can also use "bean-doctor validate_html <file> to run this manually.
    logging.info('Validating HTML output files & links.')
    files, missing, empty = scrape.validate_local_links_in_dir(output_directory)
    logging.info('Validation: %d files processed', len(files))
    for target in missing:
        logging.error("Validation error: Missing '%s'", target)
    for target in empty:
        logging.error("Validation error: Empty '%s'", target)

    # Archive if requested.
    if archival_command is not None:
        # Normalize the paths and ensure sanity before we start compression.
        output_directory = path.abspath(output_directory)
        archive_filename = path.abspath(opts.output)
        if not path.exists(output_directory):
            raise IOError("Directory to archive '{}' does not exist".format(
                output_directory))
        if path.exists(archive_filename):
            raise IOError("Output archive name '{}' already exists".format(
                archive_filename))

        # Dispatch to a particular compressor.
        if isinstance(archival_command, str):
            archive(archival_command, output_directory, archive_filename, True)
        elif callable(archival_command):
            archival_command(output_directory, archive_filename)

        # Delete the output directory.
        shutil.rmtree(output_directory)

    print("Output in '{}'".format(opts.output))


if __name__ == '__main__':
    main()
